;_________________________________________________________________
;
; This simple program demonstrates how to draw pixel graphics
; the same way it was popular in DOS programs - by directly
; writing pixels to the framebuffer memory.
;
; This way, this example can be used for learning purposes,
; instead of the outdated DOS environment.
;
; The FreshLib GUI library provides 32bit ARGB graphic.
;
; This example can be compiled for Windows and Linux
; Change the target OS from the project manager.
;_________________________________________________________________

include "%lib%/freshlib.inc"

@BinaryType GUI, compact

include "%lib%/freshlib.asm"

; Definition of the window where the drawing will happen.

iglobal
frmMainForm:
        ObjTemplate  tfEnd or tfParent, TForm, frmMain, \
                     x = 100,        \
                     y = 50,         \
                     width = 320,    \
                     height = 240,   \
                     Visible = TRUE, \
                     SplitGrid = mainSplitGrid, \
                     Caption = 'Easy drawing with FreshLib'

          ObjTemplate tfChild or tfEnd, TCanvasWindow, canvas,  \
                     x = 0, y=0, width=320, height=240,         \
                     Visible=TRUE,                              \
                     SplitCell=mainSplitGrid.canvasCell,        \
                     OnPaint = OnCanvasPaint


  SplitStart mainSplitGrid
    Split stVert or stJustGap or stOriginBR, 0, 0, 0, 0
      Cell canvasCell
      Cell nothing
  SplitEnd

endg


; The red-tape code that makes the trick works.

start:
        InitializeAll

        create  [pApplication], TApplication
        stdcall GUI.Init
        stdcall CreateFromTemplate, frmMainForm, 0
        set     [pApplication], TApplication:MainWindow, frmMain

        stdcall Run

        FinalizeAll
        stdcall TerminateAll, 0

; end of the fixed code.


; This procedure handles the painting events of the main Window.
;
; Arguments:
;            .self   - pointer to the TWindow object.
;            .pCanvas - pointer to the canvas object of the window.
;
; The canvas of the window is memory image, defined following way:
;
; struct TImage
;   .width   dd ?   ; width in pixels.
;   .height  dd ?   ; height in pixels.
;   .pPixels dd ?   ; pointer to the pixel array. Every pixel is 32 bit dword value.
; ends
;

proc OnCanvasPaint as TCanvasWindow.OnPaint
begin
        pushad
; The following code draws a chess board pattern on the whole image. Every square is 32px wide.

        mov     esi, [.pCanvas]
        mov     edi, [esi+TImage.pPixels]

        xor     edx, edx  ; Y coordinate

.loop_y:

        xor     ecx, ecx  ; X coordinate

.loop_x:

        mov     eax, ecx
        xor     eax, edx
        shr     eax, 5    ; 2^5 = 32px one field size.
        and     eax, 1    ; dark or light?

        mov     eax, [Colors+4*eax]

        mov     [edi], eax
        add     edi, 4

        inc     ecx
        cmp     ecx, [esi+TImage.width]
        jl      .loop_x

        inc     edx
        cmp     edx, [esi+TImage.height]
        jl      .loop_y

        popad
        return
endp


iglobal

  ; The array with the colors for the chess board.
  ; Notice1: These colors are with premultiplied alpha.
  ;
  ; Notice2: The semitransparent colors will be properly
  ;          displayed only if the OS allows semitransparent
  ;          windows.

  Colors dd $ff584435, $ffBD9D86
endg