module "BBCode translator"

iglobal
PHashTable tableBBtags, tpl_func,                      \
        'b',       tagStrong,                          \
        '*',       tagListItem,                        \
        'i',       tagEm,                              \
        'u',       tagUnderlined,                      \
        's',       tagDel,                             \
        'c',       tagInlineCode,                      \
        'url',     tagURL,                             \
        'img',     tagImg,                             \
        'quote',   tagQuote,                           \
        'code',    tagCode,                            \
        'size',    tagSize,                            \
        'color',   tagColor,                           \
        'list',    tagUL,                              \
        'ul',      tagUL,                              \
        'ol',      tagOL,                              \
        'table',   tagTable,                           \
        'tr',      tagTR,                              \
        'th',      tagTH,                              \
        'td',      tagTD,                              \
        'spoiler', tagSpoiler,                         \
        'email',   tagEmail,                           \
        'h1',      tagH1,                              \
        'h2',      tagH2,                              \
        'h3',      tagH3,                              \
        'h4',      tagH4,                              \
        'h5',      tagH5,                              \
        'h6',      tagH6,                              \
        ':)',      emoSmile,                           \
        ':d',      emoLol,                             \
        ':rofl:',  emoRofl,                            \
        ';)',      emoWink,                            \
        ':p',      emoTongue,                          \
        ':(',      emoSad,                             \
        ':`(',     emoCry,                             \
        '>:(',     emoAngry
endg

struct TOneTag
  .pOpenP    dd ?       ; Open tag prefix. for example '<img src="'
  .lOpenP    dd ?
  .pOpenS    dd ?       ; Open tag suffix. for example '" alt="'
  .lOpenS    dd ?
  .pClose    dd ?       ; Close tag. for example '" />'
  .lClose    dd ?
  .flags     dd ?       ; encoding flags for this tag. the first byte contain flags for the text between .pOpenP and .pOpenS
                        ; the second byte for the text between .pOpenS and .pClose
ends

; examples:
; [b]:                                  '<b>', 0, '</b>'
; [quote=title]text[/quote]:            '<blockquote><header>', '</header>', '</blockquote>'
; [img]image_url[/img]:                 '<img alt="', '" src="', '" />'
; [code=language][/code]:               '<code class="', '">', '</code>'


struc flex [arg] {
common
  local f1, f2, cnt, sz
  cnt = 0
  sz = 0
  f1 = 1
  f2 equ FALSE
forward
  match any, arg \{
    match `arg, arg \\{ f2 equ TRUE \\}
  \}

  cnt = cnt + 1
  if arg eqtype ''
    virtual at 0
      db arg
      sz = sz + $
    end virtual
  end if

common
  local lbl

  match =txt string, arg \{
    f1 = 0
    lbl text string
    f2 equ FALSE
  \}

  if (sz > 4) | (cnt>1)
    f1 = 0
  end if

  if f1
    . = arg
    .length = 0
  else
    . = lbl
    .length = lbl#\.length
  end if
  match =TRUE, f2 \{ lbl text arg \}
}


struc onetag prefix, suffix, close, flags {
common
  local ..pref, ..suff, ..close
  local ..fpr, ..fsf, ..fcl

  label . dword
if used .
  ..pref flex prefix
  ..suff flex suffix
  ..close flex close

  .pOpenP dd ..pref
  .lOpenP dd ..pref#\.length
  .pOpenS dd ..suff
  .lOpenS dd ..suff#\.length
  .pClose dd ..close
  .lClose dd ..close#\.length
  .flags  dd flags
end if
}


fEmptyAttr    = $00000001        ; empty value of the attribute is acceptable.
fBlockTag     = $00000002        ; This is a block tag. Allows paragraph generation and resets its state.
fDisableTags  = $00000004        ; Disable tag generation inside this block; The only tag that is recongized is the respective closing tag.
fLines        = $00000008        ; the tag can contain text lines.
fInlineTag    = $00000010
fMinimagBlock = $80000000        ; The tags, that are open/close by ;tagname/;end in MiniMagAsm

iglobal
  align 4

  tagStrong       onetag txt '<strong>',                              0,              txt '</strong>',    fInlineTag
  tagEm           onetag txt '<em>',                                  0,              txt '</em>',        fInlineTag
  tagUnderlined   onetag txt '<span class="underlined">',             0,              txt '</span>',      fInlineTag
  tagDel          onetag txt '<del>',                                 0,              txt '</del>',       fInlineTag
  tagInlineCode   onetag txt '<code>',                                0,              txt '</code>',      fInlineTag
  tagCode         onetag txt '<code class="block ',                   txt '">',       txt '</code>',      fBlockTag or fEmptyAttr or fDisableTags or fMinimagBlock
  tagQuote        onetag txt '<blockquote><header>',                  txt '</header>',txt '</blockquote>',fBlockTag or fEmptyAttr or fMinimagBlock
  tagUL           onetag txt '<ul>',                                  0,              txt '</ul>',        fBlockTag or fMinimagBlock
  tagOL           onetag txt '<ol type="',                            txt '">',       txt '</ol>',        fBlockTag or fEmptyAttr or fMinimagBlock
  tagListItem     onetag txt '<li>',                                  0,              txt '</li>',        fBlockTag
  tagTable        onetag txt '<table>',                               0,              txt '</table>',     fBlockTag or fMinimagBlock
  tagTR           onetag txt '<tr>',                                  0,              txt '</tr>',        fBlockTag
  tagTH           onetag txt '<th>',                                  0,              txt '</th>',        fBlockTag
  tagTD           onetag txt '<td>',                                  0,              txt '</td>',        fBlockTag
  tagSpoiler      onetag txt '<details><summary>',                    txt '</summary>',txt '</details>',   fBlockTag or fMinimagBlock
  tagURL          onetag txt '<a href="',                             txt '">',       txt '</a>',         0
  tagImg          onetag txt '<img class="block" alt="',              txt '" src="',  txt '" />',         fBlockTag  or fDisableTags
  tagInlineImg    onetag txt '<img class="inline" alt="',             txt '" src="',  txt '" />',         fInlineTag or fDisableTags
  tagSize         onetag txt '<span style="font-size:',               txt '">',       txt '</span>',      fInlineTag
  tagColor        onetag txt '<span style="color:',                   txt '">',       txt '</span>',      fInlineTag
  tagEmail        onetag txt '<a href="mailto:',                      txt '">',       txt '</a>',         0
  tagH1           onetag txt '<h1>',                                  0,              txt '</h1>',        fBlockTag or fLines
  tagH2           onetag txt '<h2>',                                  0,              txt '</h2>',        fBlockTag or fLines
  tagH3           onetag txt '<h3>',                                  0,              txt '</h3>',        fBlockTag or fLines
  tagH4           onetag txt '<h4>',                                  0,              txt '</h4>',        fBlockTag or fLines
  tagH5           onetag txt '<h5>',                                  0,              txt '</h5>',        fBlockTag or fLines
  tagH6           onetag txt '<h6>',                                  0,              txt '</h6>',        fBlockTag or fLines
  tagPara         onetag txt '<p>',                                   0,              txt '</p>',         fLines
  tagHR           onetag txt '<hr>',                                  0,              0,                  0

  tagImgMM        onetag txt '<img class="block" src="',              txt '" alt="',  txt '" />',           fBlockTag  or fDisableTags
  tagInlineImgMM  onetag txt '<img class="inline" src="',             txt '" alt="',  txt '" />',           fInlineTag or fDisableTags
  tagVideoMM      onetag txt '<video controls><source src="',         txt '">',       txt '</video>',     0

  ; Emoticons tags.

  emoSmile        onetag <'<img class="emoticon" src="', EMOTICONS_PATH, 'smile.gif" alt=":)">'>,   0,              0,              0
  emoLol          onetag <'<img class="emoticon" src="', EMOTICONS_PATH, 'lol.gif" alt=":D">'>,     0,              0,              0
  emoRofl         onetag <'<img class="emoticon" src="', EMOTICONS_PATH, 'rofl.gif" alt=":D">'>,    0,              0,              0
  emoWink         onetag <'<img class="emoticon" src="', EMOTICONS_PATH, 'wink.gif" alt=";)">'>,    0,              0,              0
  emoTongue       onetag <'<img class="emoticon" src="', EMOTICONS_PATH, 'tongue.gif" alt=":P">'>,  0,              0,              0
  emoSad          onetag <'<img class="emoticon" src="', EMOTICONS_PATH, 'sad.gif" alt=":(">'>,     0,              0,              0
  emoCry          onetag <'<img class="emoticon" src="', EMOTICONS_PATH, 'cry.gif" alt=":`(">'>,    0,              0,              0
  emoAngry        onetag <'<img class="emoticon" src="', EMOTICONS_PATH, 'angry.gif" alt=">:(">'>,  0,              0,              0

endg

; 1 - paragraph starts on every new line.
; 2 - paragraph starts on every two consecutive new lines (like in MarkDown)
NEW_PARAGRAPH_EMPTY_LINES = 1

;--------------------------------------------------------------
; Converts BB code into HTML. Starts from the position [.from]
; and ends at the end of the text or at first NULL character.
;
; Arguments:
;    .pBBCode - pointer to the TText structure with the BBCode
;               formatted text.
;    .from    - the position where to start.
;
; Returns:
;    edx - The same structure (possibly reallocated) with
;    the translated to html code text.
;

proc TranslateBBCode, .pBBCode, .from
.pStack dd ?
.ofsStart dd ?
.tagType  dd ?
.tagFlags dd ?

.fNewLine dd ?
begin
        pushad

        stdcall CreateArray, 4
        mov     [.pStack], eax

        or      [.fNewLine], -1 ; first new line implicit.
        and     [.tagFlags], 0

        mov     edx, [.pBBCode]

        OutputValue "BBCode TText.GapBegin: ", [edx+TText.GapBegin], 10, -1
        OutputValue "               GapEnd: ", [edx+TText.GapEnd], 10, -1
        OutputValue "               Length: ", [edx+TText.Length], 10, -1
        OutputValue "       Translate from: ", [.from], 10, -1

        mov     ebx, [.from]
        jmp     .loop

.next:
        inc     ebx

.loop:
        mov     ecx, [edx+TText.GapEnd]
        cmp     ebx, [edx+TText.GapBegin]
        cmovb   ecx, [edx+TText.GapBegin]
        sub     ecx, [edx+TText.GapBegin]
        add     ecx, ebx
        cmp     ecx, [edx+TText.Length]
        jae     .end_of_text

        movzx   eax, byte [edx+ecx]

        test    al, al
        jz      .end_of_text

        cmp     al, $0d
        je      .new_line

        cmp     al, $0a
        je      .new_line

        cmp     al, $20
        jbe     .next           ; skip all whitespace

        cmp     al, "["
        je      .start_tag

        cmp     [.fNewLine], NEW_PARAGRAPH_EMPTY_LINES
        mov     [.fNewLine], 0
        jb      .paragraph_ok

        test    [.tagFlags], fDisableTags
        jnz     .paragraph_ok

        push    eax
        stdcall TextMoveGap, edx, ebx
        stdcall TextSetGapSize, edx, 4
        mov     eax, [edx+TText.GapBegin]
        mov     dword [edx+eax], "<p>"
        add     [edx+TText.GapBegin], 3
        add     ebx, 3
        pop     eax

.paragraph_ok:

.continue:

; html encoding from here

        test    al, al          ; all values > 127 are unicode and should not be encoded.
        js      .next

        movzx   eax, byte [tbl_html+eax]
        test    al, al
        jz      .del_char
        jns     .next           ; the same as above

        lea     esi, [eax+tbl_html]     ; the address of the replacement string.
        lodsb
        movzx   ecx, al         ; length

; insert the replacement html encoding from esi
        stdcall TextMoveGap, edx, ebx
        stdcall TextSetGapSize, edx, ecx
        inc     [edx+TText.GapEnd]      ; delete the previous char.

        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        add     [edx+TText.GapBegin], ecx
        add     ebx, ecx

        rep movsb
        jmp     .loop

.del_char:
        stdcall TextMoveGap, edx, ebx
        inc     [edx+TText.GapEnd]      ; delete the current char.
        jmp     .loop

.not_a_tag_pop:

        add     esp, 4

.not_a_tag:

        mov     ebx, [.ofsStart]
        mov     eax, "["
        jmp     .continue

;------------------------------------------------------------------------------------

.new_line:
        inc     ebx

        mov     ecx, [edx+TText.GapEnd]
        cmp     ebx, [edx+TText.GapBegin]
        cmovb   ecx, [edx+TText.GapBegin]
        sub     ecx, [edx+TText.GapBegin]
        add     ecx, ebx
        cmp     ecx, [edx+TText.Length]
        jae     .end_of_text

        inc     [.fNewLine]

        xor     al, byte [edx+ecx]
        cmp     al, $0a xor $0d
        jne     .loop
        jmp     .next   ; skip the next char. it is part of CRLF (or LFCR sequence).

;-----------------------------------------------------------------------------

.start_tag:

        xor     ecx, ecx
        mov     [.ofsStart], ebx

.first_non_ws:
        inc     ebx

        mov     esi, [edx+TText.GapEnd]
        cmp     ebx, [edx+TText.GapBegin]
        cmovb   esi, [edx+TText.GapBegin]
        sub     esi, [edx+TText.GapBegin]
        add     esi, ebx
        cmp     esi, [edx+TText.Length]
        jae     .not_a_tag

        mov     al, [edx+esi]

        test    al, al
        jz      .not_a_tag

        cmp     al, ' '
        jbe     .first_non_ws

        mov     edi, ebx        ; the start of the hash name

        cmp     al, "/"
        je      .closing_tag

        test    [.tagFlags], fDisableTags
        jnz     .loop

.hash1:
        cmp     al, "="
        je      .tag_found
        cmp     al, "]"
        je      .tag_found

        mov     ah, al
        and     ah, $40
        shr     ah, 1
        or      al, ah  ; case insensitive hash.

        xor     cl, al
        mov     cl, [ tpl_func + ecx]

        inc     ebx

        mov     esi, [edx+TText.GapEnd]
        cmp     ebx, [edx+TText.GapBegin]
        cmovb   esi, [edx+TText.GapBegin]
        sub     esi, [edx+TText.GapBegin]
        add     esi, ebx
        cmp     esi, [edx+TText.Length]
        jae     .not_a_tag

        mov     al, [edx+esi]
        jmp     .hash1


.tag_found:

        mov     [.tagType], eax ; al = "=" or "]"

        mov     esi, [tableBBtags + sizeof.TPHashItem * ecx + TPHashItem.pKeyname]
        test    esi, esi
        jz      .not_a_tag      ; empty hash table cell.

        push    ecx

        mov     eax, ebx
        sub     eax, edi        ; the length of the tag
        movzx   ecx, byte [esi-1]  ; the length of the string in the hash table.
        cmp     ecx, eax
        jne     .not_a_tag_pop

        lea     eax, [ebx+1]
        stdcall TextMoveGap, edx, eax           ; the gap is after the tag, so, the offset and the position are the same.

        add     edi, edx           ; = pointer to the tag start. (edi contains the offset);
        xchg    esi, edi

.cmp_loop:
        dec     ecx
        js      .equal

        lodsb

        mov     ah, al
        and     ah, $40
        shr     ah, 1
        or      al, ah

        scasb
        je      .cmp_loop
        jmp     .not_a_tag_pop   ; the strings are different, so it is a collision. Ignore.

.equal:
        pop     ecx

; the tag name is the same!

        mov     eax, [tableBBtags + sizeof.TPHashItem * ecx + TPHashItem.Value]
        cmp     [eax+TOneTag.pClose], 0
        je      .stack_ok                       ; no need to push to the stack tags without closing tag.

        push    edx
        stdcall AddArrayItems, [.pStack], 1
        mov     [.pStack], edx
        mov     [eax], ecx      ; the last tag for close.
        pop     edx

.stack_ok:

        mov     edi, ebx   ; edi points one char before the attribute value or to the tag content value, depending on the tag type (with or without attribute).
        add     edi, [edx+TText.GapEnd]
        sub     edi, [edx+TText.GapBegin]

.skip_ws1:
        inc     edi
        cmp     edi, [edx+TText.Length]
        jae     .attr_here                      ; it is an empty attribute actually...

        cmp     byte [edx+edi], 0
        je      .attr_here

        cmp     byte [edx+edi], ' '
        jbe     .skip_ws1

.attr_here:
        add     edi, [edx+TText.GapBegin]
        sub     edi, [edx+TText.GapEnd]

; Here edi contains the position of the first non whitespace character of the attribute.
; Don't forget to increment it accordingly on inserting text before it.

        OutputValue "Open tag with hash=", ecx, 10, -1

        mov     eax, [edx+TText.GapBegin]
        sub     eax, [.ofsStart]

        sub     [edx+TText.GapBegin], eax       ; delete the start of the tag.
        sub     ebx, eax                        ; move ebx back, because of the deleted tag text.
        sub     edi, eax

        mov     eax, [tableBBtags + sizeof.TPHashItem * ecx + TPHashItem.Value]

        push    [eax+TOneTag.flags]
        pop     [.tagFlags]

        test    [.tagFlags], fBlockTag
        jnz     .clear_newline

        cmp     [.fNewLine], 1
        jbe     .copy_tag_prefix

        mov     esi, [tagPara.pOpenP]
        mov     ecx, [tagPara.lOpenP]
        add     edi, ecx
        call    .copy_tag

.clear_newline:

        and     [.fNewLine], 0

.copy_tag_prefix:
        mov     esi, [eax+TOneTag.pOpenP]
        mov     ecx, [eax+TOneTag.lOpenP]
        add     edi, ecx
        call    .copy_tag

        mov     esi, [eax+TOneTag.pOpenS]
        mov     ecx, [eax+TOneTag.lOpenS]
        test    esi, esi
        jz      .mono_open_tag  ; the open tag consists of only one part, without attributes.

        cmp     byte [.tagType], '='
        je      .process_attribute

        test    [eax+TOneTag.flags], fEmptyAttr         ; if the tag allows empty attribute, simply copy
        jnz     .copy                                   ; the tag suffix without trying to replace the attribute.

; here copy and encode the attribute value:
.process_attribute:
        inc     ebx

        push    esi ecx

        lea     esi, [edi-1]

.attr_loop:
        inc     esi

        mov     eax, esi
        add     eax, [edx+TText.GapEnd]
        sub     eax, [edx+TText.GapBegin]    ; eax = eax + (GapEnd - GapBegin)
        cmp     eax, [edx+TText.Length]
        jae     .end_of_attr2

        movzx   eax, byte [edx+eax]

        test    al, al
        jz      .end_of_attr2

        cmp     al, "]"
        je      .end_of_attr1

        cmp     al, "["
        je      .end_of_attr2

; encode with HTML encoding.

        test    al, al          ; all values > 127 are unicode and should not be encoded.
        js      .store_char

        movzx   eax, byte [tbl_html+eax]
        test    al, al
        jz      .attr_loop
        jns     .store_char     ; it is non-encodable character, just copy it.

        push    esi

        lea     esi, [eax+tbl_html]     ; the address of the replacement string.
        lodsb
        movzx   ecx, al         ; length

        stdcall TextSetGapSize, edx, ecx
        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        add     [edx+TText.GapBegin], ecx
        add     ebx, ecx
        add     [esp], ecx      ; increase ESI that is in the stack right now.

        rep movsb
        pop     esi
        jmp     .attr_loop

.store_char:
        stdcall TextSetGapSize, edx, 1
        mov     edi, [edx+TText.GapBegin]
        mov     [edx+edi], al
        inc     [edx+TText.GapBegin]
        inc     ebx
        inc     esi
        jmp     .attr_loop


.end_of_attr1:

        inc     esi     ; delete "]"
        dec     ebx

        add     esi, [edx+TText.GapEnd]
        sub     esi, [edx+TText.GapBegin]
        mov     [edx+TText.GapEnd], esi

.end_of_attr2:

        pop     ecx esi

.copy:
        call    .copy_tag
        test    [.tagFlags], fBlockTag
        jz      .next

        mov     [.fNewLine], NEW_PARAGRAPH_EMPTY_LINES  ; after start of a block tag, the first text starts a paragraph.
        jmp     .next


.mono_open_tag:

        cmp     byte [.tagType], '='
        jne     .next

; search the end of the attribute

        dec     edi
        add     edi, [edx+TText.GapEnd]
        sub     edi, [edx+TText.GapBegin]    ; edi = edi + (GapEnd - GapBegin)

.attr_loop2:
        inc     edi
        cmp     edi, [edx+TText.Length]
        jae     .end_of_attr3

        cmp     byte [edx+edi], 0
        je      .end_of_attr3

        cmp     byte [edx+edi], "]"
        jne     .attr_loop2

        inc     edi

.end_of_attr3:
        mov     [edx+TText.GapEnd], edi
        jmp     .next


.closing_tag:

.hash2:
        inc     ebx

        mov     esi, [edx+TText.GapEnd]
        cmp     ebx, [edx+TText.GapBegin]
        cmovb   esi, [edx+TText.GapBegin]
        sub     esi, [edx+TText.GapBegin]
        add     esi, ebx
        cmp     esi, [edx+TText.Length]
        jae     .not_a_tag

        mov     al, [edx+esi]
        test    al, al
        jz      .not_a_tag

        cmp     al, "]"
        je      .close_here

        mov     ah, al
        and     ah, $40
        shr     ah, 1
        or      al, ah  ; case insensitive hash.

        xor     cl, al
        mov     cl, [ tpl_func + ecx]
        jmp     .hash2

.close_here:

        OutputValue "Closing tag with hash=", ecx, 10, -1

        inc     edi     ; skip "/" char.

        mov     esi, [tableBBtags + sizeof.TPHashItem * ecx + TPHashItem.pKeyname]
        test    esi, esi
        jz      .not_a_tag

        test    [.tagFlags], fDisableTags
        jz      .process_closing

        mov     eax, [tableBBtags + sizeof.TPHashItem * ecx + TPHashItem.Value]
        test    [eax+TOneTag.flags], fDisableTags     ; is this the same tag, that set fDisableTags
        jz      .not_a_tag

.process_closing:
        and     [.tagFlags], 0

        push    ecx

        mov     eax, ebx
        sub     eax, edi        ; the length of the tag
        movzx   ecx, byte [esi-1]  ; the length of the string in the hash table.
        cmp     ecx, eax
        jne     .not_a_closing_tag

        lea     eax, [ebx+1]
        stdcall TextMoveGap, edx, eax           ; the gap is after the tag, so, the offset and the position are the same.

        add     edi, edx        ; the pointer to the tag start.
        xchg    esi, edi

.cmp_loop2:
        dec     ecx
        js      .equal2

        lodsb

        mov     ah, al
        and     ah, $40
        shr     ah, 1
        or      al, ah

        scasb
        je      .cmp_loop2

.not_a_closing_tag:
        pop     ecx
        jmp     .not_a_tag   ; the strings are different, so it is a collision. Ignore.

.equal2:
        pop     ecx

; the tag name is the same!

        mov     eax, [edx+TText.GapBegin]
        sub     eax, [.ofsStart]

        sub     ebx, eax                        ; move ebx back, because of the deleted tag text.
        sub     [edx+TText.GapBegin], eax       ; delete the start of the tag.

        mov     edi, [.pStack]
        mov     eax, [edi+TArray.count]
        test    eax, eax
        jz      .next           ; the stack is empty, so ignore the closing tag!

.close_loop:
        dec     eax
        js      .error_stack

        push    ecx
        mov     esi, [edi+TArray.array + 4*eax]
        mov     esi, [tableBBtags + sizeof.TPHashItem * esi + TPHashItem.Value]
        mov     ecx, [esi+TOneTag.lClose]
        mov     esi, [esi+TOneTag.pClose]
        call    .copy_tag
        pop     ecx

        cmp     ecx, [edi+TArray.array + 4*eax]
        jne     .close_loop

        mov     [edi+TArray.count], eax
        jmp     .next

.error_stack:
        inc     eax
        mov     [edi+TArray.count], eax
        jmp     .next


.end_of_text:
; here, close all open tags and free the stack array.

        stdcall TextMoveGap, edx, ebx

        mov     edi, [.pStack]
        mov     eax, [edi+TArray.count]

.closeall:
        dec     eax
        js      .finish

        mov     esi, [edi+TArray.array + 4*eax]
        mov     esi, [tableBBtags + sizeof.TPHashItem * esi + TPHashItem.Value]

        mov     ecx, [esi+TOneTag.lClose]
        mov     esi, [esi+TOneTag.pClose]
        call    .copy_tag
        jmp     .closeall

.finish:
        stdcall FreeMem, edi

        mov     [esp+4*regEDX], edx
        popad
        return


.copy_tag:
        push    edi
        stdcall TextSetGapSize, edx, ecx

        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        add     [edx+TText.GapBegin], ecx
        add     ebx, ecx

        rep movsb
        pop     edi
        retn
endp



macro HtmlEntities name, [char, encoding] {
common
  local ..tabl

  label ..tabl byte
  if used name
    label name byte
    db $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00
    db $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00
    db $20, $21, $22, $23, $24, $25, $26, $27, $28, $29, $2a, $2b, $2c, $2d, $2e, $2f
    db $30, $31, $32, $33, $34, $35, $36, $37, $38, $39, $3a, $3b, $3c, $3d, $3e, $3f
    db $40, $41, $42, $43, $44, $45, $46, $47, $48, $49, $4a, $4b, $4c, $4d, $4e, $4f
    db $50, $51, $52, $53, $54, $55, $56, $57, $58, $59, $5a, $5b, $5c, $5d, $5e, $5f
    db $60, $61, $62, $63, $64, $65, $66, $67, $68, $69, $6a, $6b, $6c, $6d, $6e, $6f
    db $70, $71, $72, $73, $74, $75, $76, $77, $78, $79, $7a, $7b, $7c, $7d, $7e, $7f

forward
  local ..offs, ..name, ..len

  if encoding eqtype ''
    ..name db ..len, encoding
    ..len = $ - ..name - 1
    ..offs = ..name - ..tabl
  else
    ..offs = encoding
  end if

  assert ..offs < 256

forward
  local ..char, ..ascii
  virtual at 0
    ..char db char
    load ..ascii byte from ..char
  end virtual
  assert ..ascii < 128
  store byte ..offs at ..tabl + ..ascii

common
  end if
  DispSize "HTML entities table:", $-..tabl
}


iglobal
  HtmlEntities tbl_html,        \
    $09, $0d,                   \
    $0a, $0a,                   \
    $0d, $0d,                   \
    '<', '&lt;',                \
    '>', '&gt;',                \
    '"', '&quot;',              \
    "'", '&apos;',              \
    '&', '&amp;'
endg





endmodule

