module "MiniMag translator"


struct TMDItem
  .type   dd ?
  .indent dd ?
ends


;--------------------------------------------------------------
; Converts MiniMag code into HTML. Starts from the position [.from]
; and ends at the end of the text or at first NULL character.
;
; Arguments:
;    .pMiniMag - pointer to the TText structure with the BBCode
;               formatted text.
;    .from    - the position where to start.
;
; Returns:
;    edx - The same structure (possibly reallocated) with
;    the translated to html code text.
;

mmfDontFormat = $80000000
mmfNoInline   = $40000000

proc TranslateMiniMag, .pText, .from
.pStack dd ?

.whspc  dd ?    ; whitespace symbols count;
.flags  dd ?
.hdr    dd ?    ; header level

.pLinks dd ?

.lbl_hash dd ?
.tmp_link dd ?

begin
        pushad

        xor     eax, eax
        mov     [.flags], eax

        stdcall GetMem, 256*4
        mov     [.pLinks], eax

        stdcall CreateArray, sizeof.TMDItem
        mov     [.pStack], eax

        mov     edx, [.pText]
        stdcall TextMoveGap, edx, [.from]

        mov     ecx, [.from]
        add     ecx, [edx+TText.GapEnd]
        sub     ecx, [edx+TText.GapBegin]
        dec     ecx

.fp_new_line:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .second_pass

        mov     al, [edx+ecx]
        test    al, al
        jz      .second_pass

        cmp     al, '['
        je      .link_def_found
        jmp     .fp_check_eol

.fp_skip_eol:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .second_pass

        mov     al, [edx+ecx]
        test    al, al
        jz      .second_pass

.fp_check_eol:
        cmp     al, $0d
        je      .fp_new_line

        cmp     al, $0a
        je      .fp_new_line

; encode HTML.

        test    al, al
        js      .fp_skip_eol

        and     eax, $ff
        movzx   eax, byte [tbl_html+eax]

        test    al, al
        jns     .fp_skip_eol

.do_encode:
        lea     esi, [eax+tbl_html]
        lodsb

        stdcall TextSetGapSize, edx, eax

        xchg    eax, ecx
        sub     eax, [edx+TText.GapEnd]
        add     eax, [edx+TText.GapBegin]
        stdcall TextMoveGap, edx, eax
        inc     [edx+TText.GapEnd]      ; delete the replaced char.

        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        add     [edx+TText.GapBegin], ecx
        rep movsb

        mov     ecx, [edx+TText.GapEnd]
        dec     ecx
        jmp     .fp_skip_eol


.link_def_found:
        mov     eax, ecx
        sub     eax, [edx+TText.GapEnd]
        add     eax, [edx+TText.GapBegin]

        stdcall TextMoveGap, edx, eax   ; move the gap to the beginning of the link.

        lea     edi, [ecx+1]  ; the start of the label.
        xor     esi, esi
        xor     eax, eax

.hash_loop:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .second_pass

        mov     al, [edx+ecx]
        test    al, al
        jz      .second_pass

        cmp     al, ']'
        je      .end_of_lbl

        test    al, al
        js      .add_hash

        mov     al, [tbl_html+eax]
        test    al, al
        jz      .fp_skip_eol    ;????????????
        jns     .add_hash

        push    ecx edx

        mov     edx, esi
        mov     esi, eax
        add     esi, tbl_html

        lodsb
        mov     ecx, eax

.html_hash:
        lodsb
        xor     edx, eax
        mov     dl, [tpl_func + edx]
        loop    .html_hash

        mov     esi, edx
        pop     edx ecx
        jmp     .hash_loop

.add_hash:
        xor     esi, eax
        movzx   esi, byte [tpl_func + esi]
        jmp     .hash_loop

.end_of_lbl:
        mov     [.lbl_hash], esi

        mov     esi, ecx

        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .second_pass

        mov     al, [edx+ecx]
        cmp     al, ' '
        jne     .fp_invalid_link

; valid label has been found...
; first: extract the label.

        sub     esi, edi
        jz      .fp_invalid_link

        cmp     esi, $ff
        ja      .fp_invalid_link

        lea     edi, [edx+edi]

        stdcall StrNew
        stdcall StrCharCat, eax, esi
        stdcall StrCatMemHTML, eax, edi, esi
        mov     [.tmp_link], eax

; fix the label length.
        stdcall StrPtr, eax
        mov     esi, eax
        mov     eax, [esi+string.len]
        dec     eax
        cmp     eax, $ff
        ja      .free_link
        mov     byte [esi], al

.link_url_start:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .free_link

        mov     al, [edx+ecx]
        test    al, al
        jz      .free_link

        cmp     al, $0d
        je      .free_link
        cmp     al, $0a
        je      .free_link

        cmp     al, ' '
        jbe     .link_url_start

        mov     edi, ecx
        mov     esi, ecx

.loop_to_end_of_link:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .free_link

        mov     al, [edx+ecx]
        test    al, al
        jz      .end_found

        cmp     al, $0d
        je      .end_of_link
        cmp     al, $0a
        je      .end_of_link

        cmp     al, ' '
        jbe     .loop_to_end_of_link

        lea     esi, [ecx+1]
        jmp     .loop_to_end_of_link


.end_of_link:
        xor     al, $0d xor $0a

        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .end_found

        cmp     al, [edx+ecx]
        jne     .end_found

        inc     ecx

.end_found:
        sub     esi, edi
        jz      .free_link

        lea     edi, [edx+edi]
        stdcall StrCatMemHTML, [.tmp_link], edi, esi

; for debugging links handing;
;        stdcall FileWriteString, [STDERR], [.tmp_link]
;        stdcall FileWriteString, [STDERR], <txt 13, 10>

        mov     eax, [.lbl_hash]
        mov     edi, [.pLinks]
        mov     esi, [.tmp_link]

.hash_search_slot:                      ; resolve possible hash collisions.
        xchg    esi, [edi+4*eax]
        test    esi, esi
        jz      .end_link

        inc     al
        cmp     eax, [.lbl_hash]
        jne     .hash_search_slot

        stdcall StrDel, esi     ; the hash table is full, so simply delete the remaining link.

.end_link:
        mov     [edx+TText.GapEnd], ecx         ; delete the entire line of the link definition.
        dec     ecx
        jmp     .fp_new_line

.free_link:
        stdcall StrDel, [.tmp_link]

.fp_invalid_link:
        mov     ecx, [edx+TText.GapEnd]
        jmp     .fp_skip_eol


; Second processing pass. It will format all the tags.

.second_pass:

        mov     ebx, [.from]

.start_of_line:

        xor     eax, eax
        mov     [.hdr], eax
        mov     byte [.flags], al
        and     [.flags], not mmfNoInline

        stdcall TextMoveGap, edx, ebx

        xor     eax, eax
        mov     [.whspc], eax

.loop:
        mov     ecx, ebx
        add     ecx, [edx+TText.GapEnd]
        sub     ecx, [edx+TText.GapBegin]
        cmp     ecx, [edx+TText.Length]
        jae     .empty_line_found

        mov     al, [edx+ecx]

        test    al, al
        jz      .empty_line_found

        cmp     al, ';'
        je      .command_found

        test    [.flags], mmfDontFormat
        jnz     .non_empty_line_found

        cmp     al, '#'
        je      .header_found

        cmp     al, '*'
        je      .list_item_found

        cmp     al, $0d
        je      .empty_line_found
        cmp     al, $0a
        je      .empty_line_found

        cmp     al, ' '
        ja      .non_empty_line_found

        inc     [.whspc]
        inc     [edx+TText.GapEnd]
        jmp     .loop


.command_found:

        cmp     [.whspc], 0
        jne     .non_empty_line_found

        test    [.flags], mmfDontFormat
        jnz     .check_for_end

        lea     eax, [ecx+8]             ; spoiler
        cmp     eax, [edx+TText.Length]
        jae     .less_than_8

        cmp     dword [edx+ecx], ';spo'
        jne     .not_spoiler
        cmp     dword [edx+ecx+4], 'iler'
        jne     .not_spoiler
        cmp     byte [edx+ecx+8], ' '
        ja      .not_spoiler

        call    .close_all_non_minimag

        push    ecx
        mov     eax, tagSpoiler
        call    .add_to_stack
        pop     ecx

        mov     eax, tagSpoiler
        call    .prefix_tag

        add     ecx, 8
        add     [edx+TText.GapEnd], 8

        push    ebx
        call    .skip_tag_arg
        pop     eax
        cmp     eax, ebx
        jne     .sum_ok

        mov     ecx, [edx+TText.GapBegin]
        mov     dword [edx+ecx], 'Spoi'
        mov     dword [edx+ecx+4], 'ler '
        add     [edx+TText.GapBegin], 7
        add     ebx, 7

.sum_ok:
        mov     eax, tagSpoiler
        call    .suffix_tag

        jmp     .start_of_line


.not_spoiler:
.less_than_8:
        lea     eax, [ecx+6]
        cmp     eax, [edx+TText.Length]
        jae     .less_than_6

        cmp     dword [edx+ecx], ';quo'
        jne     .not_quote
        cmp     word [edx+ecx+4], 'te'
        jne     .not_quote
        cmp     byte [edx+ecx+6], ' '
        ja      .not_quote

; quote command here

        call    .close_all_non_minimag

        push    ecx
        mov     eax, tagQuote
        call    .add_to_stack
        pop     ecx

        mov     eax, tagQuote
        call    .prefix_tag

        add     ecx, 6
        add     [edx+TText.GapEnd], 6

        call    .skip_tag_arg

        mov     eax, tagQuote
        call    .suffix_tag

        jmp     .start_of_line


.not_quote:

        cmp     dword [edx+ecx], ';tab'
        jne     .not_table
        cmp     word [edx+ecx+4], 'le'
        jne     .not_table
        cmp     byte [edx+ecx+6], ' '
        ja      .not_table

; table command here

        call    .close_all_non_minimag

        mov     eax, tagTable
        call    .open_tag

        jmp     .delete_to_eol


.not_table:
        cmp     dword [edx+ecx], ';beg'
        jne     .not_begin
        cmp     word [edx+ecx+4], 'in'
        jne     .not_begin
        cmp     byte [edx+ecx+6], ' '
        ja      .not_begin

        or      [.flags], mmfDontFormat

        call    .close_all_non_minimag

        push    ecx
        mov     eax, tagCode
        call    .add_to_stack
        pop     ecx

        mov     eax, tagCode
        call    .prefix_tag

        add     ecx, 6
        add     [edx+TText.GapEnd], 6

        call    .skip_tag_arg

        mov     eax, tagCode
        call    .suffix_tag

        jmp     .start_of_line


.not_begin:

        cmp     dword [edx+ecx], ';uli'
        jne     .not_ulist
        cmp     word [edx+ecx+4], 'st'
        jne     .not_ulist
        cmp     byte [edx+ecx+6], ' '
        ja      .not_ulist

        call    .close_all_non_minimag

        mov     eax, tagUL
        call    .open_tag

        jmp     .delete_to_eol


.not_ulist:
        cmp     dword [edx+ecx], ';oli'
        jne     .not_olist
        cmp     word [edx+ecx+4], 'st'
        jne     .not_olist
        cmp     byte [edx+ecx+6], ' '
        ja      .not_olist

        call    .close_all_non_minimag

        mov     eax, tagOL
        call    .open_tag

        jmp     .delete_to_eol


.not_olist:

.less_than_6:
.less_than_5:

.check_for_end:

        lea     eax, [ecx+4]
        cmp     eax, [edx+TText.Length]
        jae     .less_than_4

        cmp     dword [edx+ecx], ';end'
        jne     .not_end

        cmp     byte [edx+ecx+4], ' '
        ja      .not_end

; end command here

        call    .close_all_non_minimag
        call    .pop_one_tag

        and     [.flags], not mmfDontFormat

        jmp     .delete_to_eol


.not_end:
        test    [.flags], mmfDontFormat
        jnz     .non_empty_line_found

        cmp     dword [edx+ecx], ';---'
        jne     .not_ruler

; it is a horizontal ruller command. It can separate paragraphs by <hr> tag, or the table rows by <tr> tags.

        push    ecx
        mov     esi, [.pStack]


.row_close:
        mov     ecx, [esi+TArray.count]
        dec     ecx
        js      .put_hr ; empty stack!

        mov     eax, [esi+TArray.array + sizeof.TMDItem * ecx + TMDItem.type]

        cmp     eax, tagTD
        je      .close_this
        cmp     eax, tagTH
        je      .close_this
        cmp     eax, tagTR
        je      .close_this

        cmp     eax, tagTable
        je      .open_tr

.put_hr:
        mov     eax, tagHR
        call    .prefix_tag
        jmp     .end_hr

.close_this:
        mov     [esi+TArray.count], ecx
        call    .close_tag
        jmp     .row_close

.open_tr:
        mov     eax, tagTR
        call    .open_tag

.end_hr:
        pop     ecx
        jmp     .delete_to_eol


.less_than_4:
.not_ruler:

; it is simply a comment. Delete it up to the end of the line.

.delete_to_eol:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .eol_ok

        mov     al, [edx+ecx]
        test    al, al
        jz      .eol_ok
        cmp     al, $0d
        je      .eol_found2

        cmp     al, $0a
        jne     .delete_to_eol

.eol_found2:
        inc     ecx
        xor     al, $0d xor $0a
        cmp     al, [edx+ecx]
        jne     .eol_ok
        inc     ecx
.eol_ok:
        mov     [edx+TText.GapEnd], ecx
        jmp     .start_of_line


.list_item_found:

        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .non_empty_line_found

        cmp     byte [edx+ecx], ' '
        ja      .non_empty_line_found

        add     [edx+TText.GapEnd], 2

; here list item begins:

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]
        dec     ecx
        js      .add_leaf

        mov     eax, [esi + TArray.array + sizeof.TMDItem * ecx + TMDItem.type]
        cmp     eax, tagUL
        je      .open_li
        cmp     eax, tagOL
        jne     .add_leaf

.open_li:
        mov     eax, tagListItem
        call    .open_tag
        jmp     .add_leaf


.headers dd tagH1, tagH2, tagH3, tagH4, tagH5, tagH6

.header_found:

        xor     eax, eax

.hcnt:
        inc     eax
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .non_empty_line_found           ; not a header?

        cmp     byte [edx+ecx], '#'
        je      .hcnt

        cmp     eax, 6
        ja      .non_empty_line_found

        cmp     byte [edx+ecx], ' '
        jne     .non_empty_line_found           ; not a header?

        stc
        adc     [edx+TText.GapEnd], eax

        mov     [.hdr], eax


.non_empty_line_found:
        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]
        dec     ecx
        js      .add_leaf

        mov     eax, [esi + TArray.array + sizeof.TMDItem * ecx + TMDItem.type] ; don't add paragraphs inside elements that can handle text by themselves.
        test    [eax+TOneTag.flags], fLines
        jnz     .skip_to_eol

        cmp     eax, tagCode
        je      .skip_to_eol

        cmp     eax, tagTable
        jne     .check_for_row

        mov     eax, tagTR
        call    .open_tag
        jmp     .open_td_th

.check_for_row:
        cmp     eax, tagTR
        jne     .add_leaf

.open_td_th:
        mov     eax, tagTD
        cmp     [.hdr], 0
        je      .open_td
        mov     eax, tagTH
        and     [.hdr], 0
.open_td:
        call    .open_tag

.add_leaf:
        mov     eax, tagPara

        mov     ecx, [.hdr]
        dec     ecx
        js      .do_add_leaf
        mov     eax, [.headers + 4*ecx]

.do_add_leaf:
        call    .open_tag
        jmp     .skip_to_eol


.empty_line_found:

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]

.loop_empty_line:
        dec     ecx
        js      .skip_to_eol   ; the empty line is not inside code block because there is no blocks at all.

        mov     eax, [esi + TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        test    [eax+TOneTag.flags], fLines or fInlineTag
        jz      .check_wrap

        cmp     eax, tagInlineCode
        jne     .inline_code_ok
        and     [.flags], not mmfDontFormat
.inline_code_ok:

        mov     [esi+TArray.count], ecx ; close text leaf blocks.
        call    .close_tag
        jmp     .loop_empty_line

        dec     ecx
        js      .skip_to_eol   ; the empty line is not inside code block because there is no blocks at all.

.check_wrap:
        cmp     eax, tagTD
        je      .close_wrap
        cmp     eax, tagTH
        je      .close_wrap
        cmp     eax, tagListItem
        jne     .skip_to_eol

.close_wrap:
        mov     [esi+TArray.count], ecx
        call    .close_tag
        jmp     .skip_to_eol

; Here seems to be a good place for processing the inline formatting elements, together with the HTML encoding.

.skip_to_eol:

        xor     eax, eax
        mov     ecx, ebx
        add     ecx, [edx+TText.GapEnd]
        sub     ecx, [edx+TText.GapBegin]
        cmp     ecx, [edx+TText.Length]
        jae     .end_of_text

        mov     al, [edx+ecx]
        test    al, al
        jz      .end_of_text
        cmp     al, $0d
        je      .eol_found
        cmp     al, $0a
        je      .eol_found

        cmp     al, '`'
        je      .code

        test    [.flags], mmfDontFormat
        jnz     .next

        test    [.flags], mmfNoInline
        jnz     .inline_ok

        cmp     al, '*'
        je      .bold
        cmp     al, '/'
        je      .italic
        cmp     al, '_'
        je      .underline
        cmp     al, '-'
        je      .del

.inline_ok:
        cmp     al, '['
        je      .link

        mov     byte [.flags], al       ; the previous character.

.next:
        inc     ebx
        jmp     .skip_to_eol


.eol_found:
        mov     ah, al
        xor     ah, $0d xor $0a

        inc     ecx
        inc     ebx
        cmp     [edx+ecx], ah
        jne     .start_of_line
        inc     ebx
        jmp     .start_of_line


.link:
locals
  .linkT dd ?
  .imgW  dd ?
  .imgH  dd ?
  .linkStart dd ?
  .lblStart dd ?
  .lblLen   dd ?
  .urlLen   dd ?
  .level    dd ?
endl
        mov     ebx, ecx
        sub     ebx, [edx+TText.GapEnd]
        add     ebx, [edx+TText.GapBegin]
        stdcall TextMoveGap, edx, ebx

        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .next

        xor     eax, eax
        mov     [.linkT], eax

        mov     al, [edx+ecx]
        test    al, al
        jz      .next

        cmp     al, '!'
        je      .block_image

        cmp     al, '?'
        je      .inline_image

        cmp     al, '$'
        je      .video

        mov     [.linkT], tagURL
        jmp     .image_ok

.video:
        mov     [.linkT], tagVideoMM
        jmp     .image

.block_image:
        mov     [.linkT], tagImgMM
        jmp     .image

.inline_image:

        mov     [.linkT], tagInlineImgMM

.image:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .next

        mov     al, [edx+ecx]
        test    al, al
        jz      .next

        cmp     al, '0'
        jb      .image_ok
        cmp     al, '9'
        ja      .image_ok

; For now...
.ignore_size:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .next

        mov     al, [edx+ecx]
        test    al, al
        jz      .next

        cmp     al, ':'
        jne     .ignore_size

        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .next

        mov     al, [edx+ecx]
        test    al, al
        jz      .next

.image_ok:
; here eax contains the current byte of the address label.
; The first value of the hash is computed outside the hash loop.

        mov     edi, ecx  ; the start of the label.
        movzx   esi, byte [tpl_func + eax]    ; the initial value of the hash.

.hash_loop2:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .next

        mov     al, [edx+ecx]
        test    al, al
        jz      .next

        cmp     al, ']'
        je      .end_of_link_lbl

        xor     esi, eax
        movzx   esi, byte [tpl_func + esi]
        jmp     .hash_loop2

.end_of_link_lbl:
        mov     [.lbl_hash], esi

        mov     eax, ecx
        sub     eax, edi

        mov     [.lblStart], edi
        mov     [.lblLen], eax

; search the label in the hash table.

        push    ebx ecx

        mov     ebx, [.lbl_hash]

.lbl_search:
        mov     edi, [.pLinks]
        mov     eax, [edi+4*ebx]
        test    eax, eax
        jz      .hash_not_found

        stdcall StrPtr, eax
        mov     esi, eax
        mov     edi, edx
        add     edi, [.lblStart]
        mov     ecx, [.lblLen]

        mov     eax, [esi+string.len]
        mov     [.urlLen], eax

        movzx   eax, byte [esi] ; the first byte is the length of the label part.
        inc     esi

        sub     [.urlLen], eax
        dec     [.urlLen]

        cmp     ecx, eax
        jne     .not_this_hash

        repe cmpsb
        je      .insert_link

.not_this_hash:
        inc     bl
        cmp     ebx, [.lbl_hash]
        jne     .lbl_search

.hash_not_found:
; the link is not in the hash table, so it is a inline link.

        mov     esi, [.lblStart]
        add     esi, edx

        mov     eax, [.lblLen]
        mov     [.urlLen], eax

.insert_link:
; the label matches! Insert the tag prefix.

        pop     ecx ebx

        mov     eax, [.lblStart]
        mov     [edx+TText.GapEnd], eax ; delete the start "["

        mov     eax, [.linkT]
        call    .prefix_tag

; copy the url here. It is pointed by ESI right now.

        stdcall TextSetGapSize, edx, [.urlLen]
        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        mov     ecx, [.urlLen]
        add     ebx, ecx
        add     [edx+TText.GapBegin], ecx

        rep movsb

        mov     eax, [.linkT]
        call    .suffix_tag

; Here check for additional link content/alt content.

        mov     ecx, [edx+TText.GapEnd]
        add     ecx, [.lblLen]

        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .skip_link_content

        mov     al, [edx+ecx]
        cmp     al, '['
        jne     .skip_link_content

        lea     eax, [ecx+1]
        mov     [.lblStart], eax

        and     [.level], 0

.search_second_end:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .skip_link_content

        mov     al, [edx+ecx]
        test    al, al
        jz      .skip_link_content

        cmp     al, $0d
        je      .skip_link_content
        cmp     al, $0a
        je      .skip_link_content

        cmp     al, '['         ; nested link
        jne     .nested_ok
        inc     [.level]

.nested_ok:
        cmp     al, ']'
        jne     .search_second_end

        cmp     [.level], 0
        je      .second_end

        dec     [.level]
        jmp     .search_second_end


.second_end:
        sub     ecx, [.lblStart]
        mov     [.lblLen], ecx

        mov     ecx, [.lblStart]
        mov     [edx+TText.GapEnd], ecx ; delete the link URL/label part, because there is second one.

        push    ebx

        add     ebx, [.lblLen]
        stdcall TextMoveGap, edx, ebx
        inc     [edx+TText.GapEnd]

        mov     eax, [.linkT]
        call    .close_tag

        pop     ebx
        stdcall TextMoveGap, edx, ebx

        or      [.flags], mmfNoInline
        jmp     .skip_to_eol


.skip_link_content:

        add     ebx, [.lblLen]
        stdcall TextMoveGap, edx, ebx
        inc     [edx+TText.GapEnd]

        mov     eax, [.linkT]
        call    .close_tag

        and     [.flags], not mmfNoInline
        jmp     .skip_to_eol



.bold:
        mov     eax, tagStrong
        jmp     .inline_tag

.italic:
        mov     eax, tagEm
        jmp     .inline_tag

.underline:
        mov     eax, tagUnderlined
        jmp     .inline_tag

.del:
        mov     eax, tagDel
        jmp     .inline_tag

.code:
        mov     eax, tagInlineCode

.inline_tag:

        stdcall TextMoveGap, edx, ebx

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]
        dec     ecx
        js      .open_inline

        cmp     eax, [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        jne     .open_inline

        mov     [esi+TArray.count], ecx
        cmp     eax, tagInlineCode
        jne     @f
        and     [.flags], not mmfDontFormat
@@:
        call    .close_tag
        jmp     .end_inline

.open_inline:
        cmp     byte [.flags], ' '      ; the previous character.
        ja      .next

        mov     ecx, ebx
        add     ecx, [edx+TText.GapEnd]
        sub     ecx, [edx+TText.GapBegin]
        cmp     byte [edx+ecx+1], ' '
        jbe     .next

        cmp     eax, tagInlineCode
        jne     @f
        or      [.flags], mmfDontFormat
@@:
        call    .open_tag

.end_inline:
        inc     [edx+TText.GapEnd]
        jmp     .skip_to_eol


.end_of_text:

; Close all not closed tags:

        stdcall TextMoveGap, edx, ebx
        mov     edi, [.pStack]
        mov     ecx, [edi+TArray.count]

.close_all:
        dec     ecx
        js      .finalize

        mov     eax, [edi + TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        call    .close_tag
        jmp     .close_all

.finalize:
        mov     esi, [.pLinks]
        mov     ecx, $ff

.freelinks:
        stdcall StrDel, [esi+4*ecx]
        dec     ecx
        jns     .freelinks

        stdcall FreeMem, esi
        stdcall FreeMem, [.pStack]

        mov     [esp + 4*regEDX], edx
        popad
        return



.close_all_non_minimag:
        push    ecx esi

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]

.close_loop:
        dec     ecx
        js      .end_close

        mov     eax, [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        test    [eax+TOneTag.flags], fMinimagBlock
        jnz     .end_close

        mov     [esi+TArray.count], ecx
        call    .close_tag
        jmp     .close_loop

.end_close:
        pop     esi ecx
        retn


.pop_one_tag:
        push    ecx esi

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]
        dec     ecx
        js      .end_pop_one

        mov     [esi+TArray.count], ecx
        mov     eax, [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        call    .close_tag

.end_pop_one:
        pop     esi ecx
        retn


.open_tag:
        push    ecx esi edi
        push    eax eax
        call    .add_to_stack
        pop     eax
        call    .prefix_tag
        pop     eax
        call    .suffix_tag
        pop     edi esi ecx
        retn


.suffix_tag:
        lea     eax, [eax+TOneTag.pOpenS]
        jmp     .copy_tag

.close_tag:
        lea     eax, [eax+TOneTag.pClose]

.prefix_tag:
.copy_tag:
        push    ecx esi edi

        mov     esi, [eax]
        mov     ecx, [eax+4]
        test    esi, esi
        jz      .tag_copy_end

        add     ebx, ecx

        stdcall TextSetGapSize, edx, ecx
        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        add     [edx+TText.GapBegin], ecx

        rep movsb

.tag_copy_end:
        pop     edi esi ecx
        retn


.add_to_stack:
        push    edx eax
        stdcall AddArrayItems, [.pStack], 1
        mov     [.pStack], edx
        pop     [eax + TMDItem.type]
        pop     edx
        retn


.skip_tag_arg:
        push    eax ecx esi

.skip_tag_loop:
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .end_arg

        mov     al, [edx+ecx]
        test    al, al
        jz      .end_arg

        cmp     al, $0a
        je      .eol1
        cmp     al, $0d
        je      .eol1

        cmp     al, ' '
        jbe     .skip_tag_loop

        mov     [edx+TText.GapEnd], ecx
        mov     esi, ebx
        dec     ecx

.search_eol:
        inc     ecx
        inc     esi
        cmp     ecx, [edx+TText.Length]
        jae     .end_arg

        movzx   eax, byte [edx+ecx]
        test    al, al
        jz      .end_arg

        cmp     al, $0a
        je      .eol1

        cmp     al, $0d
        je      .eol1

        cmp     al, ' '
        cmova   ebx, esi
        jmp     .search_eol

.eol1:
        inc     ecx
        xor     al, $0d xor $0a
        cmp     al, [edx+ecx]
        jne     .end_arg
        inc     ecx
.end_arg:
        stdcall TextMoveGap, edx, ebx
        mov     [edx+TText.GapEnd], ecx

        pop     esi ecx eax
        retn

endp



; Debug procedure. Not used in production.

proc DumpStack, .pStack
begin
        pushad

        mov     esi, [.pStack]
        xor     ecx, ecx

.loop:
        cmp     ecx, [esi+TArray.count]
        jae     .finish

        mov     edi, [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type]

        stdcall FileWrite, [STDERR], [edi+TOneTag.pOpenP], [edi+TOneTag.lOpenP]
        stdcall FileWrite, [STDERR], [edi+TOneTag.pOpenS], [edi+TOneTag.lOpenS]

        stdcall FileWriteString, [STDERR], txt ": "

        stdcall NumToStr, [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.indent], ntsDec
        push    eax
        stdcall FileWriteString, [STDERR], eax
        stdcall StrDel ; from the stack

        stdcall FileWriteString, [STDERR], <txt 13, 10>

        inc     ecx
        jmp     .loop

.finish:
        popad
        return

endp



proc StrCatMemHTML, .dest, .psource, .length
begin
        pushad

        mov     esi, [.psource]
        mov     ecx, [.length]

        stdcall StrLen, [.dest]
        lea     edx, [eax+8*ecx+4]

        stdcall StrSetCapacity, [.dest], edx
        mov     edi, eax
        mov     ebx, eax
        add     edi, [edi+string.len]

        xor     eax, eax

.loop:
        dec     ecx
        js      .finish

        lodsb
        test    al, al
        js      .store

        mov     al, [tbl_html+eax]
        test    al, al
        jz      .loop
        js      .encode

.store:
        stosb
        jmp     .loop

.encode:
        push    esi ecx

        lea     esi, [eax+tbl_html]
        movzx   ecx, byte [esi]
        inc     esi

        rep movsb
        pop     ecx esi
        jmp     .loop

.finish:
        mov     ecx, edi
        xor     eax, eax
        sub     ecx, ebx
        stosd
        mov     [ebx+string.len], ecx

        popad
        return
endp







; Not finished, not working!!!

proc TranslateMarkdown, .pText, .from
.pStack dd ?

.qlvl   dd ?    ; quote level.
.whspc  dd ?    ; whitespace symbols count;
.flags  dd ?
.hdr    dd ?    ; header level

.q_from dd ?

.prev   dd ?

begin
        pushad

        stdcall CreateArray, sizeof.TMDItem
        mov     [.pStack], eax

        mov     edx, [.pText]
        mov     ebx, [.from]

.start_of_line:

        xor     eax, eax
        mov     [.qlvl], eax
        mov     [.flags], eax
        mov     [.hdr], eax

        dec     eax
        mov     [.q_from], eax          ; start from -1 on every line.

        stdcall TextMoveGap, edx, ebx

.zerospc:
        xor     eax, eax
        mov     [.whspc], eax

.loop:
        mov     ecx, ebx
        add     ecx, [edx+TText.GapEnd]
        sub     ecx, [edx+TText.GapBegin]

        mov     al, [edx+ecx]

        test    al, al
        jz      .empty_line_found

        cmp     al, '>'
        je      .quote_found

        cmp     al, '-'
        je      .ul_found
        cmp     al, '*'
        je      .ul_found
        cmp     al, '+'
        je      .ul_found

        cmp     al, '#'
        je      .header_found

        cmp     al, $0d
        je      .empty_line_found
        cmp     al, $0a
        je      .empty_line_found

        cmp     al, ' '
        ja      .non_empty_line_found

        inc     [.whspc]
        inc     [edx+TText.GapEnd]
        jmp     .loop


; closes all tags up to position ecx (excluding the tag on ecx)
.close_to_ecx:
        push    eax

.ecx_loop:
        mov     eax, [esi+TArray.count]
        dec     eax
        cmp     eax, ecx
        je      .end_close_ecx

        mov     [esi+TArray.count], eax
        mov     eax, [esi+TArray.array + sizeof.TMDItem*eax + TMDItem.type]
        call    .close_tag
        jmp     .ecx_loop

.end_close_ecx:
        pop     eax
        retn


; closes all tags up to position ecx (includig the tag on ecx)
.close_to_ecx_inc:
        push    eax

.ecx_loop2:
        mov     eax, [esi+TArray.count]
        dec     eax
        cmp     eax, ecx
        jl      .end_close_ecx

        mov     [esi+TArray.count], eax
        mov     eax, [esi+TArray.array + sizeof.TMDItem*eax + TMDItem.type]
        call    .close_tag
        jmp     .ecx_loop2


; Searches the tags stack for listitem and returns it's index in the stack.
; Arguments:
;   esi - pointer to the stack array.
; Returns:
;   esi - pointer to the tag stack.
;   If the list item has been found: CF = 0; ecx = index in the array;
;   If the list item has not been found: CF = 1
; Modifies eax

.search_for_li:

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]

.search_li:
        dec     ecx
        cmp     ecx, [.q_from]
        jle     .li_not_found

        mov     eax, [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type]

        cmp     eax, tagListItem
        je      .li_found

        cmp     eax, tagQuote
        jne     .search_li

.li_not_found:
        stc
        retn

.li_found:
        clc
        retn


.search_for_bq:

        mov     esi, [.pStack]
        mov     ecx, [.q_from]

.search_bq:
        inc     ecx
        cmp     ecx, [esi+TArray.count]
        jae     .bq_not_found

        cmp     [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type], tagQuote
        jne     .search_bq

        clc
        retn

.bq_not_found:
        stc
        retn


.ul_found:
        call    .check_for_code_block
        jc      .skip_to_eol

        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .empty_line_found

        cmp     byte [edx+ecx], ' '     ; mandatory space
        ja      .non_empty_line_found

        add     [edx+TText.GapEnd], 2   ; remove the marker.

.search_list:

        call    .search_for_li
        jnc     .list_found

        call    .close_to_ecx
        jmp     .create_new_ul


.list_found:

        OutputValue "List found at: ", [.whspc], 10, -1

        mov     edi, [.whspc]
        sub     edi, [esi + TArray.array + sizeof.TMDItem*ecx + TMDItem.indent]

        call    .close_to_ecx

        cmp     edi, 4
        jge     .create_new_ul  ; create list one level of nesting up.

        mov     [esi + TArray.count], ecx  ; close the current listitem.
        call    .close_tag

        cmp     edi, -4
        jg      .create_new_li

        dec     ecx

        mov     eax, [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type]             ; eax should be tagUL or tagOL and nothing else!
        mov     [esi+TArray.count], ecx
        call    .close_tag             ; close the list as well. (can be UL or OL!)

        jmp     .search_list           ; and then search back for the list on the previous level.


.create_new_ul:

        mov     eax, tagUL
        call    .open_tag

.create_new_li:
        OutputValue "List item created on:", [.whspc], 10, -1

        mov     eax, tagListItem
        call    .open_tag
        jmp     .loop           ; after creating list items, the whitespace for the current line should not be reset.


.check_for_code_block:

        mov     esi, [.pStack]
        mov     eax, [esi+TArray.count]
        dec     eax
        js      .not_a_code_block

        cmp     [esi+TArray.array + sizeof.TMDItem*eax + TMDItem.type], tagCode
        jne     .not_a_code_block

; there is open code block
        cmp     [.whspc], 4
        jb      .close_code_block

        stc
        retn

.close_code_block:

        mov     eax, tagCode
        call    .close_tag

.not_a_code_block:
        clc
        retn


.quote_found:

        DebugMsg "Quote found"

        mov     eax, [.q_from]
        OutputValue "Search quote from: ", eax, 10, -1

.quote_search:
        inc     eax
        mov     [.q_from], eax
        cmp     eax, [esi+TArray.count]
        jae     .create_new_quote

        cmp     [esi+TArray.array+sizeof.TMDItem*eax+TMDItem.type], tagQuote
        je      .del_marker
        jmp     .quote_search

.create_new_quote:

        DebugMsg "New quote created"

        call    .check_for_code_block
        jc      .skip_to_eol

        mov     eax, tagQuote
        call    .open_tag

        mov     eax, [esi+TArray.count]
        dec     eax
        mov     [.q_from], eax


.del_marker:

        DebugMsg "Delete quote marker."
        inc     [edx+TText.GapEnd]

        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .empty_line_found

        cmp     byte [edx+ecx], ' '     ; optional space.
        jne     .zerospc

        inc     [edx+TText.GapEnd]
        jmp     .zerospc




.headers dd tagH1, tagH2, tagH3, tagH4, tagH5, tagH6

.header_found:

        cmp     [.whspc], 4
        jae     .code_block

        xor     eax, eax

.hcnt:
        inc     eax
        inc     ecx
        cmp     ecx, [edx+TText.Length]
        jae     .non_empty_line_found           ; not a header?

        cmp     byte [edx+ecx], '#'
        je      .hcnt

        cmp     eax, 6
        ja      .non_empty_line_found

        cmp     byte [edx+ecx], ' '
        jne     .non_empty_line_found           ; not a header?

        stc
        adc     [edx+TText.GapEnd], eax

        mov     [.hdr], eax

.non_empty_line_found:

        OutputValue "Non empty line indent: ", [.whspc], 10, -1

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]
        dec     ecx
        js      .add_leaf

        mov     ecx, [esi + TArray.array + sizeof.TMDItem * ecx + TMDItem.type] ; don't add paragraphs inside elements that can handle text by themselves.
        test    [ecx+TOneTag.flags], fLines
        jnz     .skip_to_eol

.add_leaf:
        call    .search_for_bq
        jc      .quote_ok

        call    .close_to_ecx_inc

.quote_ok:

        call    .search_for_li
        mov     eax, [.whspc]

        jc      .indent_ok
        sub     eax, [esi + TArray.array + sizeof.TMDItem*ecx+TMDItem.indent]
.indent_ok:

        cmp     eax, 4
        jge     .code_block

        cmp     eax, -4
        jg      .para

        call    .close_to_ecx_inc

        dec     ecx
        mov     eax, [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        mov     [esi+TArray.count], ecx
        call    .close_tag

        jmp     .add_leaf

.para:
        mov     eax, tagPara
        mov     ecx, [.hdr]
        dec     ecx
        js      .do_add_leaf
        mov     eax, [.headers + 4*ecx]

.do_add_leaf:
        call    .open_tag
        jmp     .skip_to_eol


.empty_line_found:

        call    .search_for_bq
        jc      .quote_ok2

        call    .close_to_ecx_inc

.quote_ok2:

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]
        dec     ecx
        js      .skip_to_eol   ; the empty line is not inside code block because there is no blocks at all.

        mov     eax, [esi + TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        test    [eax+TOneTag.flags], fLines
        jz      .close_list

        mov     [esi+TArray.count], ecx ; close text leaf blocks.
        call    .close_tag
        jmp     .skip_to_eol

.close_list:
        DebugMsg "Close list on second empty line."

        cmp     eax, tagListItem
        jne     .skip_to_eol

        call    .close_tag

        mov     [esi+TArray.count], ecx
        dec     ecx
        js      .skip_to_eol

        mov     eax, [esi + TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        cmp     eax, tagUL
        jne     .skip_to_eol

        call    .close_tag
        mov     [esi+TArray.count], ecx
        jmp     .skip_to_eol


.code_block:

        DebugMsg "Code block detected!"

        mov     esi, [.pStack]
        mov     ecx, [esi+TArray.count]
        dec     ecx
        js      .add_new_code_block

        cmp     [esi+TArray.array + sizeof.TMDItem*ecx + TMDItem.type], tagCode
        je      .skip_to_eol

.add_new_code_block:

        mov     eax, tagCode
        call    .open_tag

;        add     [edx+TText.GapEnd], 4   ; the gap should be properly positioned here!
        jmp     .skip_to_eol


.open_new:
        cmp     ecx, [esi + TArray.count]       ; esi = indicators list
        jae     .skip_to_eol

        mov     eax, [esi + TArray.array + sizeof.TMDItem*ecx + TMDItem.indent]
        mov     [.whspc], eax
        mov     eax, [esi + TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        call    .open_tag

        inc     ecx
        jmp     .open_new

; Here seems to be a good place for processing the inline formatting elements, together with the HTML encoding.

.skip_to_eol:

        xor     eax, eax
        mov     ecx, ebx
        add     ecx, [edx+TText.GapEnd]
        sub     ecx, [edx+TText.GapBegin]
        cmp     ecx, [edx+TText.Length]
        jae     .end_of_text

        mov     al, [edx+ecx]
        test    al, al
        jz      .end_of_text
        cmp     al, $0d
        je      .eol_found
        cmp     al, $0a
        je      .eol_found

; html encoding from here

        test    al, al          ; all values > 127 are unicode and should not be encoded.
        js      .next

        movzx   eax, byte [tbl_html+eax]
        test    al, al
        jns     .next           ; the same as above

        lea     esi, [eax+tbl_html]     ; the address of the replacement string.
        lodsb
        movzx   ecx, al         ; length

; insert the replacement html encoding from esi

        stdcall TextSetGapSize, edx, ecx
        stdcall TextMoveGap, edx, ebx
        inc     [edx+TText.GapEnd]      ; delete the previous char.

        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        add     [edx+TText.GapBegin], ecx
        add     ebx, ecx

        rep movsb

.next:
        inc     ebx
        jmp     .skip_to_eol


.eol_found:
        mov     ah, al
        xor     ah, $0d xor $0a

        inc     ecx
        inc     ebx
        cmp     [edx+ecx], ah
        jne     .start_of_line
        inc     ebx
        jmp     .start_of_line

.end_of_text:

; Close all not closed tags:

        stdcall TextMoveGap, edx, ebx
        mov     edi, [.pStack]
        mov     ecx, [edi+TArray.count]

.close_all:
        dec     ecx
        js      .finalize

        mov     eax, [edi + TArray.array + sizeof.TMDItem*ecx + TMDItem.type]
        call    .close_tag
        jmp     .close_all

.finalize:
        stdcall FreeMem, [.pStack]
;        stdcall FreeMem, [.pIList]

        mov     [esp + 4*regEDX], edx
        popad
        return


.open_tag:
        push    ecx esi edi
        push    eax eax
        call    .add_to_stack
        pop     eax
        call    .prefix_tag
        pop     eax
        call    .suffix_tag
        pop     edi esi ecx
        retn


.suffix_tag:
        lea     eax, [eax+TOneTag.pOpenS]
        jmp     .copy_tag

.close_tag:
        lea     eax, [eax+TOneTag.pClose]

.prefix_tag:
.copy_tag:
        push    ecx esi edi

        mov     esi, [eax]
        mov     ecx, [eax+4]
        test    esi, esi
        jz      .tag_copy_end

        add     ebx, ecx

        stdcall TextSetGapSize, edx, ecx
        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        add     [edx+TText.GapBegin], ecx

        rep movsb

.tag_copy_end:
        pop     edi esi ecx
        retn


.add_to_stack:
        push    edx eax [.whspc]
        stdcall AddArrayItems, [.pStack], 1
        mov     [.pStack], edx
        pop     [eax+TMDItem.indent]
        pop     [eax + TMDItem.type]
        pop     edx
        retn

endp






endmodule





