; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: OS independent part of different random number generators.
;
;  Target OS: Any
;
;  Dependencies: memory.asm, strlib.asm
;
;  Notes:
;
;_________________________________________________________________________________________

module "Random library"


; Cryptographically strong random generator. Possibly slow. OS dependent implementation.
; Returns:
;   eax - pointer to memory block of random bytes.

interface GetRandomBytes, .count




proc GetRandomString, .size
begin
        pushad

        mov     eax, [.size]
        shr     eax, 1
        adc     eax, 0
        jz      .error

        stdcall GetRandomBytes, eax
        jc      .error

        mov     esi, eax
        mov     ebx, eax

        stdcall StrNew
        mov     [esp+4*regEAX], eax

        stdcall StrSetCapacity, eax, [.size]
        mov     edi, eax
        mov     ecx, [.size]
        mov     [edi+string.len], ecx

        xor     eax, eax

.loop:
        lodsb
        mov     dl, al
        and     al, $0f
        shr     dl, 4

        cmp     al,$0a
        sbb     al,$69
        das

        stosb

        dec     ecx
        jz      .string_ok

        mov     al, dl
        cmp     al,$0a
        sbb     al,$69
        das

        stosb
        dec     ecx
        jnz     .loop

.string_ok:
        xor     eax, eax
        stosd
        stdcall FreeMem, ebx

        clc
        popad
        return

.error:
        stc
        popad
        return
endp



include '%TargetOS%/random.asm'



endmodule