


; Computes the SHA-1 hash for the data in [.pData] with length in  [.len]
; The memory buffer in [.pData] must have size higher than [.len] and multiply of 64 bytes (512 bits)
;
;

proc DataSHA1, .pData, .len
.hash  rd 5     ; the h[0]..h[4] accumulators
.abcde rd 5     ; the chunk hashes
.words rd 80    ; the 80 dword chunks.
begin
        pushad

; init the accumulators
        mov     ecx, 4

.init_acc:
        mov     eax, [.sha1_h + 4*ecx]
        mov     [.hash+4*ecx], eax
        mov     [.abcde+4*ecx], eax
        dec     ecx
        jns     .init_acc

; pad the data

        mov     edx, [.len]

        mov     edi, [.pData]
        add     edi, edx
        lea     ecx, [edx+1]
        mov     al, $80
        stosb

        add     edx, $40
        and     edx, not $3f
        sub     ecx, edx
        neg     ecx

        sub     ecx, 8
        jge     @f
        add     ecx, $40
        add     edx, $40
@@:
        xor     eax, eax
        rep stosb
        stosd

        mov     eax, [.len]
        shl     eax, 3          ; in bits!!!
        bswap   eax
        stosd

; edx contains the padded length of the block. multiply of 64

        shr     edx, 2  ; the size in dwords
        mov     esi, [.pData]

.chunk_loop:

        xor     ebx, ebx

.get16dw:
        lodsd
        bswap   eax
        mov     [.words + 4*ebx], eax

        inc     ebx
        cmp     ebx, 16
        jne     .get16dw

.fill80:
        mov     eax, [.words + 4*ebx - 3*4]
        xor     eax, [.words + 4*ebx - 8*4]
        xor     eax, [.words + 4*ebx - 14*4]
        xor     eax, [.words + 4*ebx - 16*4]
        rol     eax, 1
        mov     [.words + 4*ebx], eax

        inc     ebx
        cmp     ebx, 80
        jne     .fill80

; process one chunk.

        xor     ebx, ebx

.process0_19:
        mov     ecx, [.abcde + 1*4]       ; b
        mov     eax, ecx
        not     ecx                     ; not b

        and     eax, [.abcde + 2*4]       ; c
        and     ecx, [.abcde + 3*4]       ; d
        or      ecx, eax                ; f

        mov     eax, [.abcde + 0*4]
        rol     eax, 5                  ; a rol 5
        lea     eax, [eax + ecx + $5A827999]; f + k
        add     eax, [.abcde + 4*4]       ; e
        add     eax, [.words + 4*ebx]     ; temp

        mov     ecx, [.abcde + 1*4]
        rol     ecx, 30
        push    eax            [.abcde + 0*4] ecx            [.abcde + 2*4] [.abcde + 3*4]
        pop     [.abcde + 4*4] [.abcde + 3*4] [.abcde + 2*4] [.abcde + 1*4] [.abcde + 0*4]

        inc     ebx
        cmp     ebx, 20
        jne     .process0_19

.process20_39:

        mov     ecx, [.abcde + 1*4]
        xor     ecx, [.abcde + 2*4]
        xor     ecx, [.abcde + 3*4]       ; f

        mov     eax, [.abcde + 0*4]
        rol     eax, 5
        lea     eax, [eax + ecx + $6ED9EBA1]
        add     eax, [.abcde + 4*4]
        add     eax, [.words + 4*ebx]     ; temp

        mov     ecx, [.abcde + 1*4]
        rol     ecx, 30
        push    eax [.abcde + 0*4] ecx [.abcde + 2*4] [.abcde + 3*4]
        pop     [.abcde + 4*4] [.abcde + 3*4] [.abcde + 2*4] [.abcde + 1*4] [.abcde + 0*4]

        inc     ebx
        cmp     ebx, 40
        jne     .process20_39


.process40_59:

        mov     ecx, [.abcde+1*4]       ; b
        mov     eax, ecx                ; b
        mov     edi, [.abcde+2*4]       ; c

        and     ecx, edi                ; b and c
        and     eax, [.abcde+3*4]       ; b and d
        and     edi, [.abcde+3*4]       ; c and d

        or      ecx, eax
        or      ecx, edi                ; f

        mov     eax, [.abcde+0*4]
        rol     eax, 5
        lea     eax, [eax+ecx+$8F1BBCDC]
        add     eax, [.abcde+4*4]
        add     eax, [.words+4*ebx]     ; temp

        mov     ecx, [.abcde+1*4]
        rol     ecx, 30
        push    eax [.abcde+0*4] ecx [.abcde+2*4] [.abcde+3*4]
        pop     [.abcde+4*4] [.abcde+3*4] [.abcde+2*4] [.abcde+1*4] [.abcde+0*4]

        inc     ebx
        cmp     ebx, 60
        jne     .process40_59

.process60_79:
        mov     ecx, [.abcde+1*4]
        xor     ecx, [.abcde+2*4]
        xor     ecx, [.abcde+3*4]       ; f

        mov     eax, [.abcde+0*4]
        rol     eax, 5
        lea     eax, [eax+ecx+$CA62C1D6]
        add     eax, [.abcde+4*4]
        add     eax, [.words+4*ebx]     ; temp

        mov     ecx, [.abcde+1*4]
        rol     ecx, 30
        push    eax [.abcde+0*4] ecx [.abcde+2*4] [.abcde+3*4]
        pop     [.abcde+4*4] [.abcde+3*4] [.abcde+2*4] [.abcde+1*4] [.abcde+0*4]

        inc     ebx
        cmp     ebx, 80
        jne     .process60_79

; the word array is processed

        mov     ecx, 4

.add_loop:
        mov     eax, [.abcde+4*ecx]
        add     eax, [.hash+4*ecx]
        mov     [.abcde+4*ecx], eax
        mov     [.hash+4*ecx], eax
        dec     ecx
        jns     .add_loop

        sub     edx, 16
        jnz     .chunk_loop


        stdcall GetMem, 20
        mov     edi, eax

        mov     eax, [.hash+0*4]
        mov     ebx, [.hash+1*4]
        mov     ecx, [.hash+2*4]
        mov     edx, [.hash+3*4]
        mov     esi, [.hash+4*4]

        bswap   eax
        bswap   ebx
        bswap   ecx
        bswap   edx
        bswap   esi

        mov     [edi], eax
        mov     [edi+4], ebx
        mov     [edi+8], ecx
        mov     [edi+12], edx
        mov     [edi+16], esi

        mov     [esp+4*regEAX], edi
        popad
        return

.sha1_h dd 0x67452301, 0xEFCDAB89, 0x98BADCFE, 0x10325476, 0xC3D2E1F0

endp



proc StrSHA1, .pSha1
begin
        push    ebx ecx esi

        mov     esi, [.pSha1]
        stdcall StrNew
        mov     ebx, eax
        mov     ecx, 20

.loop:
        movzx   eax, byte [esi]
        inc     esi

        stdcall NumToStr, eax, ntsHex or ntsFixedWidth or ntsUnsigned + 2
        stdcall StrCat, ebx, eax
        stdcall StrDel, eax
        dec     ecx
        jnz     .loop

        mov     eax, ebx

        pop     esi ecx ebx
        return
endp


