; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Memory based images manipulation library.
;
;  Target OS: Win32
;
;  Dependencies: memory.asm
;
;  Notes:
;_________________________________________________________________________________________

uses msimg32, gdi32


struct TWinImage
  . TImage
; os dependent data
  .hBitmap dd ?
ends



body CreateImage
  .bmi BITMAPV4HEADER
begin
        pushad

        stdcall GetMem, sizeof.TWinImage
        jc      .finish

        mov     esi, eax

        xor     eax, eax
        lea     edi, [.bmi]
        mov     ecx, sizeof.BITMAPV4HEADER/4
        rep stosd

        inc     eax

        mov     ecx, [.width]
        mov     edx, [.height]

        cmp     ecx, 0
        cmovle  ecx, eax

        cmp     edx, 0
        cmovle  edx, eax

        mov     [esi+TWinImage.width], ecx
        mov     [esi+TWinImage.height], edx
        mov     [esi+TWinImage.wrapW], ecx
        mov     [esi+TWinImage.wrapH], edx

        mov     [.bmi.biSize], sizeof.BITMAPV4HEADER
        mov     [.bmi.biPlanes], 1

; it is already 0
;        mov     [.bmi.biCompression], BI_RGB
;        mov     [.bmi.biSizeImage], 0
;        mov     [.bmi.biClrUsed], 0
;        mov     [.bmi.biClrImportant], 0

        mov     [.bmi.BlueMask],  $000000ff
        mov     [.bmi.GreenMask], $0000ff00
        mov     [.bmi.RedMask],   $00ff0000
        mov     [.bmi.AlphaMask], $ff000000

        mov     [.bmi.biBitCount], 32
        neg     edx
        mov     [.bmi.biWidth], ecx
        mov     [.bmi.biHeight], edx

        lea     eax, [.bmi]
        lea     ecx, [esi+TWinImage.pPixels]
        xor     edx, edx
        invoke  CreateDIBSection, edx, eax, DIB_RGB_COLORS, ecx, edx, edx
        mov     [esi+TWinImage.hBitmap], eax

        mov     [esp+4*regEAX], esi

        lea     esi, [esi+TWinImage.lock]
        stdcall MutexCreate, 0, esi
        stdcall MutexRelease, esi

        clc

.finish:
        popad
        return
endp





body DestroyImage
begin
        pushad

        mov     esi, [.ptrImage]
        test    esi, esi
        jz      .finish

.lock_wait:
        stdcall LockImg, esi
        jc      .lock_wait


        invoke  GdiFlush
        invoke  DeleteObject, [esi+TWinImage.hBitmap]

        lea     edi, [esi+TWinImage.lock]
        stdcall MutexDestroy, edi

        stdcall FreeMem, esi

.finish:
        popad
        return
endp



;;AC_SRC_OVER =  $00000000
;;AC_SRC_ALPHA = $01000000
;;
;;struct BLENDFUNCTION
;;  .BlendOp             db ?
;;  .BlendFlags          db ?
;;  .SourceConstantAlpha db ?
;;  .AlphaFormat         db ?
;;ends
;
;body DrawImageRect
;begin
;        push    eax ecx edx esi edi
;
;        mov     esi, [.pImage]
;        test    esi, esi
;        jz      .exit
;
;        stdcall LockImg, esi
;        jc      .exit
;
;        invoke  CreateCompatibleDC, 0
;        mov     ebx, eax
;
;        invoke  SelectObject, ebx, [esi+TWinImage.hBitmap]
;        push    eax
;
;        invoke  BitBlt, [.where], [.xDst], [.yDst], [.width], [.height], ebx, [.xSrc], [.ySrc], SRCCOPY
;
;        invoke  SelectObject, ebx ; from the stack
;        invoke  DeleteDC, ebx
;
;        stdcall UnlockImg, esi
;
;.exit:
;        pop     edi esi edx ecx eax
;        return
;endp
;


body DrawImageRect

.maskUL RECT ; upper left
.maskUR RECT ; upper right
.maskLL RECT ; lower left
.maskLR RECT ; lower right

.sRect  RECT ; the rectangle used for intersecting the source rectangle.

.dc         dd ?
.oldbmp     dd ?

begin
        pushad

        mov     esi, [.pImage]
        test    esi, esi
        jz      .finish

        stdcall LockImg, esi
        jc      .finish

        xor     eax, eax
        mov     ecx, [esi+TImage.wrapW]
        mov     edx, [esi+TImage.wrapH]

        mov     [.maskUL.left], eax     ; UL
        mov     [.maskUL.top], eax      ; UL
        mov     [.maskLL.left], eax             ; LL
        mov     [.maskUR.top], eax                      ; UR

        mov     [.maskUL.right], ecx    ; UL
        mov     [.maskUL.bottom], edx   ; UL
        mov     [.maskLR.left], ecx                             ; LR
        mov     [.maskLR.top], edx                              ; LR
        mov     [.maskLL.top], edx              ; LL
        mov     [.maskLL.right], ecx            ; LL
        mov     [.maskUR.left], ecx                     ; UR
        mov     [.maskUR.bottom], edx                   ; UR

        shl     ecx, 1
        shl     edx, 1

        mov     [.maskUR.right], ecx                    ; UR
        mov     [.maskLL.bottom], edx           ; LL
        mov     [.maskLR.right], ecx                            ; LR
        mov     [.maskLR.bottom], edx                           ; LR

        mov     ecx, [.xSrc]
        mov     edx, [.ySrc]
        mov     [.sRect.left], ecx
        mov     [.sRect.top], edx
        add     ecx, [.width]
        add     edx, [.height]
        mov     [.sRect.right], ecx
        mov     [.sRect.bottom], edx

        lea     eax, [.sRect]
        lea     edx, [.maskUL]
        stdcall RectIntersect, eax, eax, edx
        jc      .exit

; at least some part of the image, has to be drawn.

        invoke  CreateCompatibleDC, 0
        mov     [.dc], eax

        invoke  SelectObject, [.dc], [esi+TWinImage.hBitmap]
        mov     [.oldbmp], eax

        mov     ebx, [esi+TImage.orgX]
        mov     edx, [esi+TImage.orgY]

        add     [.sRect.left], ebx
        add     [.sRect.top], edx
        add     [.sRect.right], ebx
        add     [.sRect.bottom], edx

        mov     ebx, [esi+TImage.wrapW]
        mov     edx, [esi+TImage.wrapH]
        lea     ecx, [.sRect]

        lea     eax, [.maskUL]
        stdcall RectIntersect2, eax, ecx
        jc      .upper_left_ok

        call    .DoDraw                                 ; the upper left rect is always drawn

.upper_left_ok:
        lea     eax, [.maskUR]
        stdcall RectIntersect2, eax, ecx
        jc      .upper_right_ok

        sub     [.maskUR.left], ebx
        sub     [.maskUR.right], ebx

        mov     edi, [.maskUL.right]    ; width
        sub     edi, [.maskUL.left]
        add     [.xDst], edi

        call    .DoDraw

        sub     [.xDst], edi


.upper_right_ok:
        lea     eax, [.maskLL]
        stdcall RectIntersect2, eax, ecx
        jc      .lower_left_ok

        sub     [.maskLL.top], edx
        sub     [.maskLL.bottom], edx

        mov     edi, [.maskUR.bottom]
        sub     edi, [.maskUR.top]
        add     [.yDst], edi

        call    .DoDraw

        sub     [.yDst], edi

.lower_left_ok:
        lea     eax, [.maskLR]
        stdcall RectIntersect2, eax, ecx
        jc      .lower_right_ok

        sub     [.maskLR.left], ebx
        sub     [.maskLR.top], edx
        sub     [.maskLR.right], ebx
        sub     [.maskLR.bottom], edx

        mov     edi, [.maskUL.right]
        sub     edi, [.maskUL.left]
        add     [.xDst], edi
        mov     edi, [.maskUR.bottom]
        sub     edi, [.maskUR.top]
        add     [.yDst], edi

        call    .DoDraw

.lower_right_ok:

        invoke  SelectObject, [.dc], [.oldbmp]
        invoke  DeleteDC, [.dc]

.exit:
        stdcall UnlockImg, esi

.finish:
        popad
        return


; Draws the rectangle pointed by EAX

.DoDraw:
        pushad

; computes in right/bottom, width/height
        mov     ebx, [eax+RECT.bottom]
        mov     edx, [eax+RECT.right]
        sub     ebx, [eax+RECT.top]
        sub     edx, [eax+RECT.left]

; draws
        invoke  BitBlt, [.where], [.xDst], [.yDst], edx, ebx, [.dc], [eax+RECT.left], [eax+RECT.top], SRCCOPY

        popad
        retn

endp
