module "TCanvasWindow"


object TCanvasWindow, TWindow
  ._canvas        dd ?    ; TImage containing the image of the window only.
  ._canvas_valid  dd ?
  ._OnPaint       dd ?

  param .Canvas, ._canvas, NONE
  param .OnPaint, ._OnPaint, .SetOnPaint


  method .SetOnPaint, .value

  method .ForceRefresh
  method .__RenderView
  method .__UpdateImages
  method .SelfPaint, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height    ; Paints itself on the caller provided TImage.

endobj


interface TCanvasWindow.OnPaint, .self, .pCanvas



method TCanvasWindow.ForceRefresh
begin
        mov     eax, [.self]
        and     [eax+TCanvasWindow._canvas_valid], 0
        inherited
        return
endp


method TCanvasWindow.SelfPaint       ; .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height
begin
        pushad
        mov     esi, [.self]

        cmp     [esi+TCanvasWindow._canvas_valid], 0
        jne     @f
        exec    esi, TCanvasWindow:__RenderView
@@:
        mov     ebx, [esi+TCanvasWindow._canvas]
        test    ebx, ebx
        jz      .finish

        stdcall BlendImage, [.pDstImage], [.xDst], [.yDst], ebx, [.xSrc], [.ySrc], [.width], [.height]

.finish:
        popad
        inherited [.pDstImage], [.xDst], [.yDst], [.xSrc], [.ySrc], [.width], [.height]
        return
endp




method TCanvasWindow.__UpdateImages
begin
        pushad
        mov     esi, [.self]

;        DebugMsg "TCanvasWindow update canvas."

        mov     ebx, [esi+TCanvasWindow._visible]
        mov     edi, [esi+TCanvasWindow._canvas]

        mov     ecx, [esi+TCanvasWindow._width]
        mov     edx, [esi+TCanvasWindow._height]

        test    ebx, ebx
        jz      .destroy

        test    edi, edi
        jz      .create_new

; check the size

        cmp     ecx, [edi+TImage.width]
        ja      .destroy

        cmp     edx, [edi+TImage.height]
        jbe     .setwrap

.destroy:
;        DebugMsg "Destroy old canvas."

        stdcall DestroyImage, edi                         ; canvas

        xor     eax, eax
        mov     [esi+TCanvasWindow._canvas], eax

        test    ebx, ebx
        jz      .exit


.create_new:

;        DebugMsg "Create new canvas."

; new size - get it with x1.5 reserve in order to allow easy resizing.
;        lea     ecx, [ecx*3]
;        lea     edx, [edx*3]
;        shr     ecx, 1
;        shr     edx, 1

; The above reserve is too big. Lets make it more RAM-friendly.

        lea     eax, [ecx+256]
        lea     ebx, [edx+256]

        stdcall CreateImage, eax, ebx
        mov     [esi+TCanvasWindow._canvas], eax
        mov     edi, eax


;        OutputValue "Canvas created: ", eax, 16, 8

.setwrap:
        mov     [edi+TImage.wrapW], ecx
        mov     [edi+TImage.wrapH], edx

.exit:
        and     [esi+TCanvasWindow._canvas_valid], 0
        popad
        inherited
        return
endp




method TCanvasWindow.__RenderView
begin
        pushad

        mov     esi, [.self]

        get     eax, esi, TCanvasWindow:OnPaint
        test    eax, eax
        jz      .user_paint_ok

        get     ecx, esi, TCanvasWindow:Canvas
        test    ecx, ecx
        jz      .user_paint_ok

        stdcall eax, esi, ecx                   ; user OnPaint handler.

.user_paint_ok:
        inc     [esi+TCanvasWindow._canvas_valid]
        popad
        return
endp



method TCanvasWindow.SetOnPaint
begin
        push    eax esi

        mov     esi, [.self]
        mov     eax, [.value]
        cmp     eax, [esi+TCanvasWindow._OnPaint]
        je      .finish

        mov     [esi+TCanvasWindow._OnPaint], eax
        exec    esi, TWindow:RectChanged2, 0

.finish:
        pop     esi eax
        return
endp

;_________________________________________________________________________________________



endmodule