; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TMenu object class
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: Represents main menu or popup menu objects.
;_________________________________________________________________________________________
module "TMenu library"


mitSeparator = 0
mitAction    = 1
mitSubmenu   = 2
mitString    = 3        ; contains simple text string can't have accelerator and icons.



struct TMenuItem
  .type   dd ?
  .item   dd ?    ; different for the different types of items: NULL, pointer to TAction, pointer to TMenu, handle of string.
  .itemID dd ?    ; user assigned ID
  .height dd ?
ends


mitEnd = $80


macro MenuTemplate name {

  macro action actvar, id \{
    \local ..sz, ..beg
    ..beg = $
    db ..sz
    db mitAction
    dd actvar
    if ~id eq
      dd id
    else
      dd 0
    end if
    ..sz = $-..beg
  \}

  macro submenu text, id \{
    \local ..sz, ..beg, ..lbl
    ..beg = $
    db ..sz
    db mitSubmenu
    if ~id eq
      dd id
    else
      dd 0
    end if
    ..lbl db text, 0
    ..sz = $-..beg
  \}

  macro text txt, id \{
    \local ..sz, ..beg, ..lbl
    ..beg = $
    db ..sz
    db mitString
    if ~id eq
      dd id
    else
      dd 0
    end if
    ..lbl db txt, 0
    ..sz = $-..beg
  \}

  macro separator \{
    db 2
    db mitSeparator
  \}

  macro endsub \{
    db 2
    db mitEnd
  \}

  label name byte
}


macro EndMenu {
  endsub
  purge action
  purge submenu
  purge text
  purge separator
  purge endsub
}



object TMenu, TCanvasWindow
  ._p_items   dd ?
  ._selected  dd ?
  .__caller   dd ?

  .__icon_size dd ?

  ._ActiveSubmenu dd ?

  ._onstring dd ?

  param .OnString, ._onstring, ._onstring

  param .Selected, ._selected, .SetSelected
  param .Items, ._p_items, .SetItems

  method .SetSelected, .value
  method .SetItems, .menu_template
  method .SetVisible, .value

  method .AddItem, .Type, .Item, .itemID
  method .InsertItem, .Type, .Item, .itemID, .position

  method .__RenderView

  method .Create, .Parent
  method .Destroy

  method .Show, .parent, .x, .y

  method .EventButtonPress, .button, .kbdState, .x, .y
  method .EventMouseMove, .x, .y, .kbdState

endobj


interface TMenu.OnString, .self, .item




method TMenu.Create
begin
        stdcall CreateArray, sizeof.TMenuItem
        push    eax

        mov     eax, [.self]
        pop     [eax+TMenu._p_items]

        mov     [eax+TMenu.__want_focus], 0
        mov     [eax+TMenu._selected], -1

        inherited 0

        set     [.self], TWindow:border, borderNone
        return
endp




method TMenu.Destroy
begin
        mov     eax, [.self]
        stdcall FreeMem, [eax+TMenu._p_items]
        return
endp



method TMenu.SetItems
begin
        pushad




        popad
        return
endp


method TMenu.AddItem
begin
        pushad

        mov     esi, [.self]
        mov     edx, [esi+TMenu._p_items]

        stdcall AddArrayItems, edx, 1
        mov     [esi+TMenu._p_items], edx

        pushd   [.Type]
        pushd   [.Item]
        pushd   [.itemID]
        popd    [eax+TMenuItem.itemID]
        popd    [eax+TMenuItem.item]
        popd    [eax+TMenuItem.type]

        mov     [eax+TMenuItem.height], 10

        popad
        return
endp


method TMenu.InsertItem
begin
        pushad

        mov     esi, [.self]
        mov     edx, [esi+TMenu._p_items]

        stdcall InsertArrayItems, edx, [.position], 1
        mov     [esi+TMenu._p_items], edx

        pushd   [.Type]
        pushd   [.Item]
        pushd   [.itemID]
        popd    [eax+TMenuItem.itemID]
        popd    [eax+TMenuItem.item]
        popd    [eax+TMenuItem.type]

        mov     [eax+TMenuItem.height], 10

        popad
        return
endp


method TMenu.SetVisible ;

.maxW dd ?
.sumH dd ?
.str  dd ?

begin
        pushad

        mov     esi, [.self]

        cmp     [.value], FALSE
        je      .switch_off

        get     edi, esi, TMenu:Items

        mov     ecx, [edi+TArray.count]
        test    ecx, ecx
        jnz     .compute_width

        stdcall GetTextBounds, "Empty menu", 10, [GUI.DefaultFont]
        add     eax, 20
        jmp     .set_width


.compute_width:
; first scan for the icon height

        lea     edx, [edi+TArray.array-sizeof.TMenuItem]
        mov     [esi+TMenu.__icon_size], 0

.loop_icon:
        add     edx, sizeof.TMenuItem
        dec     ecx
        js      .height_ok

        cmp     [edx+TMenuItem.type], mitAction
        jne     .loop_icon

        get     eax, [edx+TMenuItem.item], TAction:Icon
        test    eax, eax
        jz      .loop_icon

        mov     eax, [eax+TImage.height]
        cmp     eax, [esi+TMenu.__icon_size]
        cmovb   eax, [esi+TMenu.__icon_size]
        mov     [esi+TMenu.__icon_size], eax
        jmp     .loop_icon

.height_ok:
        mov     ecx, [edi+TArray.count]
        lea     edi, [edi+TArray.array-sizeof.TMenuItem]        ; edi points to the current TMenuItem

        xor     eax, eax
        mov     [.maxW], eax                    ; the width of the menu window
        mov     [.sumH], eax                    ; the total height of the menu window

.loop:
        add     edi, sizeof.TMenuItem
        dec     ecx
        js      .end_width

        cmp     [edi+TMenuItem.type], mitSeparator
        je      .separator

        cmp     [edi+TMenuItem.type], mitAction
        je      .action

        cmp     [edi+TMenuItem.type], mitSubmenu
        je      .submenu

; it is string.

        stdcall StrDup, [edi+TMenuItem.item]
        push    eax
        jmp     .str_width

.separator:

        mov     edx, [GUI.menuSeparatorHeight]
        jmp     .set_height

.submenu:
        get     eax, [edi+TMenuItem.item], TMenu:Caption
        push    eax

        stdcall StrCat, eax, txt "  >"
        jmp     .str_width


.action:
        get     eax, [edi+TMenuItem.item], TAction:Accelerator
        push    eax

        get     edx, [edi+TMenuItem.item], TAction:Caption
        stdcall StrCat, eax, edx


.str_width:
        stdcall StrPtr, eax
        stdcall GetTextBounds, eax, [eax+string.len], [GUI.DefaultFont]

        stdcall StrDel ; from the stack

        cmp     eax, [.maxW]
        jbe     .comp_height

        mov     [.maxW], eax

.comp_height:

        mov     eax, [esi+TMenu.__icon_size]
        add     eax, [GUI.menuIconMargin]
        add     edx, [GUI.menuIconMargin]       ; edx = height

        cmp     edx, eax
        cmovb   edx, eax

.set_height:
        mov     [edi+TMenuItem.height], edx
        add     [.sumH], edx

        jmp     .loop

.end_width:

; here compute the size of the menu and set .width and .height fields.

        mov     edx, [GUI.iconMenuChecked]
        mov     eax, [edx+TImage.width]
        add     eax, [esi+TMenu.__icon_size]

        add     eax, [.maxW]
        add     eax, [GUI.menuMinTextDist]

        mov     edx, [GUI.menuIconMargin]
        lea     edx, [3*edx]
        add     eax, edx
        mov     edx, [GUI.boxBorderWidth]
        lea     eax, [eax+2*edx]


.set_width:
        set     esi, TWindow:width, eax

        mov     edx, [.sumH]
        mov     eax, [GUI.boxBorderWidth]
        lea     edx, [edx+2*eax]

        mov     eax, [GUI.menuIconMargin]
        lea     edx, [edx+2*eax]

        set     esi, TWindow:height, edx

.finish:
        popad

        inherited [.value]
        return


.switch_off:
        xor     ecx, ecx

        mov     eax, [esi+TMenu.__caller]
        test    eax, eax
        jz      .set_active

        istype  eax, TMenu
        cmove   ecx, eax
        je      .set_active

;.remove_capture:
        stdcall SetMouseCapture, ecx    ; ecx is 0 here!!!

.set_active:
        mov     [__ActiveMenu], ecx
        jmp     .finish

endp



method TMenu.__RenderView

.bounds TBounds

.y    dd ?
.ytxt dd ?
.xTxt dd ?
.clTxt dd ?

.txtline dd ?
.ascender dd ?

.count dd ?

.canvas dd ?

; item parameters, depend on the item type.

.text    dd ?
.accel   dd ?
.icon    dd ?
.check   dd ?
.enabled dd ?
.subicon dd ?

begin
        pushad
        mov     esi, [.self]

;        exec    esi, TMenu:__UpdateCanvas

        mov     ebx, [esi+TMenu._canvas]
        mov     [.canvas], ebx
        test    ebx, ebx
        jz      .end_loop

        xor     eax, eax
        mov     [.bounds.x], eax
        mov     [.bounds.y], eax

        mov     eax, [esi+TWindow._width]
        mov     ecx, [esi+TWindow._height]
        mov     [.bounds.width], eax
        mov     [.bounds.height], ecx

        lea     eax, [.bounds]
        stdcall [DrawBox], [.canvas], eax, [GUI.clMenuBack], [GUI.borderMenu], [GUI.boxBorderWidth]

        get     edi, esi, TMenu:Items
        test    edi, edi
        jz      .end_loop


; compute the text base line.

        stdcall GetFontMetrics, [GUI.DefaultFont]
        mov     [.txtline], eax
        lea     ebx, [2*ebx]
        mov     [.ascender], ebx

        mov     ecx, [GUI.boxBorderWidth]
        add     ecx, [GUI.menuIconMargin]
        mov     [.y], ecx

        mov     ecx, [GUI.menuIconMargin]
        lea     ecx, [3*ecx]
        mov     eax, [esi+TMenu.__icon_size]

        add     ecx, eax
        add     ecx, [GUI.boxBorderWidth]

        mov     eax, [GUI.iconMenuChecked]
        add     ecx, [eax+TImage.width]

        mov     [.xTxt], ecx

        mov     ecx, [edi+TArray.count]
        lea     edi, [edi+TArray.array]

        mov     [.count], ecx

        xor     ecx, ecx
        dec     ecx

.loop:
        inc     ecx
        cmp     ecx, [.count]
        jae     .end_loop

        mov     eax, [edi+TMenuItem.type]

        cmp     eax, mitSeparator
        je      .draw_separator

        cmp     eax, mitAction
        je      .action

        cmp     eax, mitSubmenu
        je      .submenu

        cmp     eax, mitString
        jne     .next_item

.string:
        stdcall StrDup, [edi+TMenuItem.item]
        mov     [.text], eax

        xor     eax, eax
        mov     [.accel], eax
        mov     [.icon], eax
        mov     [.check], eax
        mov     [.subicon], eax
        inc     eax
        mov     [.enabled], eax
        jmp     .draw_background

.submenu:
        get     eax, [edi+TMenuItem.item], TMenu:Caption
        mov     [.text], eax

        get     eax, [edi+TMenuItem.item], TMenu:Enabled
        mov     [.enabled], eax

        mov     ebx, [GUI.menuSubIconGray]
        test    eax, eax
        jz      @f

        mov     ebx, [GUI.menuSubIcon]

        cmp     ecx, [esi+TMenu._selected]
        jne     @f
        mov     ebx, [GUI.menuSubIconSel]

@@:
        mov     [.subicon], ebx

        xor     eax, eax
        mov     [.icon], eax
        mov     [.check], eax
        mov     [.accel], eax
        jmp     .draw_background


.action:
        mov     [.subicon], 0

        get     eax, [edi+TMenuItem.item], TAction:Caption
        mov     [.text], eax

        get     eax, [edi+TMenuItem.item], TAction:Accelerator
        mov     [.accel], eax

        get     eax, [edi+TMenuItem.item], TAction:Enabled
        mov     [.enabled], eax

        get     eax, [edi+TMenuItem.item], TAction:Icon
        mov     [.icon], eax

        get     eax, [edi+TMenuItem.item], TAction:Checked
        mov     [.check], eax
        jmp     .draw_background


.draw_separator:
        mov     eax, [.bounds.x]
        mov     ebx, [.bounds.width]
        mov     edx, [GUI.menuIconMargin]
        add     eax, edx
        sub     ebx, edx
        sub     ebx, edx

        mov     edx, [edi+TMenuItem.height]
        add     edx, [.y]
        add     edx, [.y]
        sar     edx, 1

        stdcall DrawSolidRect, [.canvas], eax, edx, ebx, 1, [GUI.clBorderDark]
        inc     edx
        stdcall DrawSolidRect, [.canvas], eax, edx, ebx, 1, [GUI.clBorderLight]
        jmp     .next_item


; draw the background
.draw_background:
        cmp     ecx, [esi+TMenu._selected]
        jne     .selected_ok

        cmp     [.enabled], 0
        je      .selected_ok

        mov     eax, [.bounds.x]
        mov     edx, [.bounds.width]

        add     eax, [GUI.boxBorderWidth]
        sub     edx, [GUI.boxBorderWidth]
        sub     edx, [GUI.boxBorderWidth]

        stdcall DrawSolidRect, [.canvas], eax, [.y], edx, [edi+TMenuItem.height], [GUI.clMenuSelection]

        mov     ebx, [.icon]
        test    ebx, ebx
        jz      .selected_ok

        mov     edx, [ebx+TImage.width]
        add     edx, [GUI.menuIconMargin]

        mov     ebx, [GUI.iconMenuChecked]
        add     eax, [ebx+TImage.width]

        mov     ebx, [GUI.menuIconMargin]
        add     eax, ebx
        shr     ebx, 1
        add     eax, ebx

        stdcall DrawSolidRect, [.canvas], eax, [.y], edx, [edi+TMenuItem.height], [GUI.clMenuSelIcon]

.selected_ok:

; draw the icon

        cmp     [.icon], 0
        je      .icon_ok

        mov     ebx, [GUI.iconMenuChecked]
        mov     eax, [ebx+TImage.width]

        add     eax, [GUI.menuIconMargin]
        add     eax, [GUI.menuIconMargin]
        add     eax, [GUI.boxBorderWidth]       ; x-coordinate

        mov     ebx, [.icon]
        test    ebx, ebx
        jz      .icon_ok

        mov     edx, [edi+TMenuItem.height]
        sub     edx, [ebx+TImage.height]
        add     edx, [.y]
        add     edx, [.y]
        sar     edx, 1

        stdcall BlendImage, [.canvas], eax, edx, ebx, 0, 0, [ebx+TImage.width], [ebx+TImage.height]

.icon_ok:

; draw the checkbox

        cmp     [.check], 0
        je      .draw_text

        mov     eax, [GUI.menuIconMargin]
        add     eax, [GUI.boxBorderWidth]

        mov     ebx, [GUI.iconMenuChecked]

        cmp     [.enabled], 0
        jne     .check_ok

        mov     ebx, [GUI.iconMenuCheckedGray]

.check_ok:
        mov     edx, [edi+TMenuItem.height]
        sub     edx, [ebx+TImage.height]
        add     edx, [.y]
        add     edx, [.y]
        sar     edx, 1

        stdcall BlendImage, [.canvas], eax, edx, ebx, 0, 0, [ebx+TImage.width], [ebx+TImage.height]


.draw_text:

        mov     eax, [GUI.clMenuTextGray]

        cmp     [.enabled], 0
        je      .txt_color_ok

        mov     eax, [GUI.clMenuText]

        cmp     ecx, [esi+TMenu._selected]
        jne     .txt_color_ok

        mov     eax, [GUI.clEditSelTxt]

.txt_color_ok:
        mov     [.clTxt], eax

        push    eax
        push    [GUI.DefaultFont]

        mov     eax, [edi+TMenuItem.height]
        sub     eax, [.txtline]
        add     eax, [.ascender]
        sar     eax, 1
        add     eax, [.y]
        mov     [.ytxt], eax

        push    eax
        pushd   [.xTxt]

        stdcall StrLen, [.text]
        push    eax

        stdcall StrPtr, [.text]

        stdcall DrawString, [.canvas], eax    ; remaining from the stack

        stdcall StrDel, [.text]

; draw the submenu arrow
        mov     ebx, [.subicon]
        test    ebx, ebx
        jz      .draw_accelerator

        mov     eax, [.bounds.width]
        sub     eax, [ebx+TImage.width]
        sub     eax, [GUI.menuIconMargin]

        mov     edx, [edi+TMenuItem.height]
        sub     edx, [ebx+TImage.height]
        add     edx, [.y]
        add     edx, [.y]
        sar     edx, 1

        stdcall BlendImage, [.canvas], eax, edx, ebx, 0, 0, [ebx+TImage.width], [ebx+TImage.height]
        jmp     .next_item

; draw the accelerator
.draw_accelerator:
        cmp     [.accel], 0
        je      .next_item

        stdcall StrPtr, [.accel]
        push    eax

        stdcall GetTextBounds, eax, [eax+string.len], [GUI.DefaultFont]

        mov     edx, [.bounds.width]
        sub     edx, [GUI.boxBorderWidth]
        sub     edx, [GUI.menuIconMargin]
        sub     edx, eax

        pop     eax
        stdcall DrawString, [.canvas], eax, [eax+string.len], edx, [.ytxt], [GUI.DefaultFont], [.clTxt]

        stdcall StrDel, [.accel]


.next_item:
        mov     eax, [edi+TMenuItem.height]
        add     [.y], eax

        add     edi, sizeof.TMenuItem
        jmp     .loop


.end_loop:

        popad
        return

.empty_menu:
        stdcall DrawString, [.canvas], "Empty menu", 10, 10, 14, [GUI.DefaultFont], [GUI.clMenuText]
        stdcall FilterDisabled, [.canvas], FALSE
        jmp     .end_loop

endp




method TMenu.Show
begin
        pushad

        mov     esi, [.self]
        mov     edi, [.parent]

        exec    edi, TWindow:ClientToScreenXY, [.x], [.y]

        mov     [esi+TMenu._x], ecx
        mov     [esi+TMenu._y], edx

        mov     [esi+TMenu.__caller], edi
        istype  edi, TMenu
        jne     .submenu_ok

        mov     eax, esi
        xchg    eax, [edi+TMenu._ActiveSubmenu]

        test    eax, eax
        jz      .submenu_ok

        set     eax, TMenu:Visible, FALSE
        stdcall SetMouseCapture, edi

.submenu_ok:
        mov     [__ActiveMenu], esi

        set     esi, TMenu:Visible, TRUE
        set     esi, TMenu:Selected, 0
        popad
        return
endp




proc AcceleratorToStr, .pFastKey
begin
        mov     eax, [.pFastKey]
        test    eax, eax
        jz      .empty

        stdcall CreateKeyName, [eax+TFastKey.key], [eax+TFastKey.flags]
        jnc     .finish

.empty:
        stdcall StrNew
        stc

.finish:
        return
endp


; Converts the string representation of an accelerator key to
; TFastKey structure.
;
; Returns:
;   eax - keyboard flags (TFastKey.flags)
;   edx - keyboard key (TFastKey.key)
;
; ToDo: more modifiers need to be processed

proc StrToAccelerator, .hString
.res TFastKey
begin
        pushad

        stdcall StrSplitList, [.hString], "+", FALSE
        mov     edx, eax
        xor     ecx, ecx
        mov     [.res.flags], ecx
        mov     [.res.key], ecx
        dec     ecx

.loop:
        inc     ecx
        cmp     ecx, [edx+TArray.count]
        jae     .finish

        mov     ebx, Ctrl
        stdcall StrCompNoCase, [edx+TArray.array + 4*ecx], txt "Ctrl"
        jc      .setflag
        stdcall StrCompNoCase, [edx+TArray.array + 4*ecx], txt "Control"
        jc      .setflag
        stdcall StrCompNoCase, [edx+TArray.array + 4*ecx], txt "Ctl"
        jc      .setflag

        mov     ebx, Alt
        stdcall StrCompNoCase, [edx+TArray.array + 4*ecx], txt "Alt"
        jc      .setflag

        mov     ebx, Shift
        stdcall StrCompNoCase, [edx+TArray.array + 4*ecx], txt "Shift"
        jc      .setflag

        stdcall KbdNameToCode, [edx+TArray.array + 4*ecx]
        jc      .loop
        mov     [.res.key], eax
        jmp     .loop

.setflag:
        or      [.res.flags], ebx
        jmp     .loop

.finish:
        stdcall ListFree, edx, StrDel
        mov     eax, [.res.flags]
        mov     edx, [.res.key]
        mov     [esp+4*regEAX], eax
        mov     [esp+4*regEDX], edx
        popad
        return
endp




method TMenu.SetSelected
begin
        pushad

        mov     esi, [.self]
        mov     ecx, [.value]
        cmp     ecx, [esi+TMenu._selected]
        je      .finish

        get     edi, esi, TMenu:Items
        test    edi, edi
        jz      .finish

        mov     eax, [esi+TMenu._selected]
        mov     [esi+TMenu._selected], ecx

        cmp     eax, [edi+TArray.count]
        jae     .old_ok

        imul    eax, sizeof.TMenuItem
        cmp     [edi + TArray.array + eax + TMenuItem.type], mitSubmenu
        jne     .old_ok

        set     [edi + TArray.array + eax + TMenuItem.item], TMenu:Visible, FALSE

.old_ok:
        cmp     ecx, [edi+TArray.count]
        jae     .finish

; compute Y coordinate

        lea     edi, [edi+TArray.array]
        xor     ebx, ebx

.loop:
        dec     ecx
        js      .height_ok

        add     ebx, [edi+TMenuItem.height]
        add     edi, sizeof.TMenuItem
        jmp     .loop

.height_ok:

        mov     eax, ebx
        add     eax, [edi+TMenuItem.height]
        exec    esi, TMenu:RectChanged2, 0;, ebx, [esi+TWindow._width], eax

        cmp     [edi+TMenuItem.type], mitSubmenu
        jne     .hide_old

        get     eax, [edi+TMenuItem.item], TMenu:Enabled
        test    eax, eax
        jz      .hide_old

        mov     eax, [esi+TMenu._width]
        sub     eax, [GUI.menuIconMargin]

        exec    [edi+TMenuItem.item], TMenu:Show, esi, eax, ebx ; will hide the previous open submenu!
        jmp     .finish

; hide old menu if active:
.hide_old:
        cmp     [esi+TMenu._ActiveSubmenu], 0
        je      .finish

        set     [esi+TMenu._ActiveSubmenu], TMenu:Visible, FALSE
        and     [esi+TMenu._ActiveSubmenu], 0

.finish:
        popad
        return
endp



method TMenu.EventMouseMove
begin
        pushad

        mov     esi, [.self]

        mov     ecx, [.x]
        mov     edx, [.y]
        sub     ecx, [esi+TMenu._x]
        js      .check_caller
        sub     edx, [esi+TMenu._y]
        js      .check_caller

        cmp     ecx, [esi+TMenu._width]
        jge     .check_caller

        cmp     edx, [esi+TMenu._height]
        jge      .check_caller

        sub     edx, [GUI.boxBorderWidth]
        js      .exit

        sub     edx, [GUI.menuIconMargin]
        js      .exit

; search for item.

        get     edi, esi, TMenu:Items

        mov     ecx, [edi+TArray.count]
        lea     edi, [edi+TArray.array]

        xor     ebx, ebx        ; index
        xor     eax, eax        ; Y coordinate

.loop:
        dec     ecx
        js      .exit

        add     eax, [edi+TMenuItem.height]
        cmp     edx, eax
        jl      .found

        inc     ebx
        add     edi, sizeof.TMenuItem
        jmp     .loop

.found:
        set     esi, TMenu:Selected, ebx

.exit:
        popad
        return

.check_caller:
        mov     esi, [esi+TMenu.__caller]
        test    esi, esi
        jz      .exit

        istype  esi, TMenu
        jne     .exit

        exec    esi, TMenu:EventMouseMove, [.x], [.y], [.kbdState]
        jmp     .exit
endp





method TMenu.EventButtonPress
begin
        pushad
        mov     esi, [.self]

        mov     ecx, [.x]
        mov     edx, [.y]

        sub     ecx, [esi+TMenu._x]
        js      .check_caller
        sub     edx, [esi+TMenu._y]
        js      .check_caller
        cmp     ecx, [esi+TMenu._width]
        jge     .check_caller
        cmp     edx, [esi+TMenu._height]
        jge     .check_caller

        mov     ecx, [esi+TMenu._selected]
        mov     edx, [esi+TMenu._p_items]

        cmp     ecx, [edx+TArray.count]
        jae     .finish

        mov     eax, sizeof.TMenuItem
        imul    ecx, eax

        lea     edi, [edx+TArray.array+ecx]

        cmp     [edi+TMenuItem.type], mitSeparator
        je      .finish

        cmp     [edi+TMenuItem.type], mitString
        je      .onstring

        cmp     [edi+TMenuItem.type], mitSubmenu
        je      .finish

        cmp     [edi+TMenuItem.type], mitAction
        je      .onaction

        jmp     .finish

.onstring:
        call    .hideall

        get     eax, esi, TMenu:OnString
        test    eax, eax
        jz      .finish

        stdcall eax, esi, edi
        jmp     .finish

.onaction:
        get     eax, [edi+TMenuItem.item], TAction:Enabled
        test    eax, eax
        jz      .finish

        call    .hideall
        exec    [edi+TMenuItem.item], TAction:Execute, [.self]   ; execute the action.

.finish:
        popad
        return


.check_caller:
        set     esi, TMenu:Visible, FALSE       ; hide the menu

        mov     esi, [esi+TMenu.__caller]
        test    esi, esi
        jz      .finish

        istype  esi, TMenu
        jne     .finish

        mov     [esi+TMenu._selected], -1
        exec    esi, TMenu:EventMouseMove, [.x], [.y], [.kbdState]
        exec    esi, TMenu:EventButtonPress, [.button], [.kbdState], [.x], [.y] ; what about converting coordinates to the caller ???
        jmp     .finish


.hideall:
        push    esi

.hide_loop:
        set     esi, TMenu:Visible, FALSE
        mov     esi, [esi+TMenu.__caller]
        test    esi, esi
        jz      .end_hide

        istype  esi, TMenu
        jnz     .end_hide

        set     esi, TMenu:Selected, -1
        jmp     .hide_loop

.end_hide:
        pop     esi
        retn
endp





proc ___MenuFromTemplate
begin
        push    ebx ecx edx edi

        create  ebx, TMenu, 0

.loop:
        movzx   edx, byte [esi+1]       ; the type of the item

        cmp     edx, mitEnd
        je      .end_of_menu

        cmp     edx, mitSeparator
        je      .separator

        cmp     edx, mitAction
        je      .action

        cmp     edx, mitString
        je      .string

        cmp     edx, mitSubmenu
        je      .submenu

        jne     .next_item

.separator:

        xor     ecx, ecx
        xor     edi, edi
        jmp     .add_item

.action:
        mov     edi, [esi+6]    ; the item ID
        mov     ecx, [esi+2]
        mov     ecx, [ecx]
        jmp     .add_item

.string:
        mov     edi, [esi+2]    ; the item ID
        lea     ecx, [esi+6]
        stdcall StrDupMem, ecx
        mov     ecx, eax
        jmp     .add_item

.submenu:
        mov     edi, [esi+2]    ; the item ID
        lea     ecx, [esi+6]
        movzx   eax, byte [esi]
        add     esi, eax

        call    ___MenuFromTemplate
        set     eax, TMenu:Caption, ecx
        set     eax, TMenu:Enabled, TRUE
        set     eax, TMenu:Visible, FALSE
        mov     ecx, eax

.add_item:
        exec    ebx, TMenu:AddItem, edx, ecx, edi

.next_item:
        movzx   eax, byte [esi]
        add     esi, eax
        jmp     .loop

.end_of_menu:
        mov     eax, ebx
        pop     edi edx ecx ebx
        return
endp



proc MenuFromTemplate, .pTemplate
begin
        push    esi
        mov     esi, [.pTemplate]
        call    ___MenuFromTemplate
        pop     esi
        return
endp



endmodule


