; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TScrollWindow object class
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: TScrollWindow is window that may have scrollers to scroll the client area.
;_________________________________________________________________________________________
module "TScrollable library"


struct TScroller
  .pos   dd ?
  .page  dd ?
  .max   dd ?
  .state dd ?
ends

; Constants for TScroller.state - the state of the particular scrollbar.

sbsNormal = 0   ; when the mouse is outside the scrollbar
sbsActive = 1   ; when the mouse points on the scrollbar

; Constants for TScrollable._state - the state of the window as a whole.

scsReleased = 0           ; no scrollbar is in drag state.
scsDragHorizontal = 1     ; the horizontal scrollbar is in drag state.
scsDragVertical   = 2     ; the vertical scrollbar is in drag state.



object TScrollable, TWindow

  ._scrX  TScroller
  ._scrY  TScroller

  ._drag_ofs dd ?

  ._state    dd ?

  ._imgX     dd ?
  ._imgY     dd ?

  param .PosX,  ._scrX.pos,  .SetPos:2
  param .PageX, ._scrX.page, .SetPage:2
  param .MaxX,  ._scrX.max,  .SetMax:2

  param .PosY,  ._scrY.pos,  .SetPos:2
  param .PageY, ._scrY.page, .SetPage:2
  param .MaxY,  ._scrY.max,  .SetMax:2

  method .SetPos,  .param, .value
  method .SetPage, .param, .value
  method .SetMax,  .param, .value

  method .__UpdateImages
  method .__DrawScrollbar, .scrollbar   ; .scrollbar == TScrollable._scrX or TScrollable._scrY
  method .__Invalidate, .scrollbar
  method .SelfPaint, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height    ; Paints itself on the caller provided TImage.

  method .EventMouseMove, .x, .y, .kbdState
  method .EventMouseLeave

  method .EventButtonPress, .button, .kbdState, .x, .y
  method .EventButtonRelease, .button, .kbdState, .x, .y
  method .EventScroll, .direction, .command, .value

endobj


;_________________________________________________________________________________________

method TScrollable.__Invalidate ; .scrollbar
begin
        pushad
        mov     esi, [.self]

        cmp     [.scrollbar], TScrollable._scrY
        je      .invalidateY

        xor     eax, eax
        mov     edx, [GUI.scrollWidth]
        mov     ebx, [esi+TScrollable._height]
        mov     ecx, [esi+TScrollable._width]
        sub     ebx, edx
        jmp     .invalidate

.invalidateY:

        xor     ebx, ebx
        mov     ecx, [GUI.scrollWidth]

        mov     eax, [esi+TScrollable._width]
        mov     edx, [esi+TScrollable._height]
        sub     eax, ecx

.invalidate:

        exec    esi, TScrollable:RectChangedXY2, eax, ebx, ecx, edx
        popad
        return
endp

;_________________________________________________________________________________________


method TScrollable.__UpdateImages
begin
        inherited
        pushad

        mov     esi, [.self]

        mov     edi, [esi+TScrollable._imgX]
        mov     ecx, [esi+TScrollable._width]
        mov     edx, [GUI.scrollWidth]

        test    edi, edi
        jz      .createx

        cmp     [edi+TImage.wrapW], ecx
        jne     .destroyx
        cmp     [edi+TImage.wrapH], edx
        je      .ximg_ok

.destroyx:

        stdcall DestroyImage, edi

.createx:
        stdcall CreateImage, ecx, edx
        mov     [esi+TScrollable._imgX], eax

.ximg_ok:

        mov     edi, [esi+TScrollable._imgY]
        mov     ecx, [esi+TScrollable._height]
        mov     edx, [GUI.scrollWidth]

        test    edi, edi
        jz      .createy

        cmp     [edi+TImage.wrapH], ecx
        jne     .destroyy
        cmp     [edi+TImage.wrapW], edx
        je      .yimg_ok

.destroyy:

        stdcall DestroyImage, edi

.createy:
        stdcall CreateImage, edx, ecx
        mov     [esi+TScrollable._imgY], eax

.yimg_ok:

        exec    esi, TScrollable:__DrawScrollbar, TScrollable._scrX
        exec    esi, TScrollable:__DrawScrollbar, TScrollable._scrY
        popad
        return
endp


;_________________________________________________________________________________________


method TScrollable.SetPos       ;, .param, .value
begin
        pushad

        mov     esi, [.self]
        mov     eax, [.value]

        mov     edx, TScrollable._scrX
        mov     ecx, TScrollable._scrY
        cmp     [.param], TScrollable.PosX
        cmovne  edx, ecx

        cmp     eax, [esi+edx+TScroller.pos]
        je      .finish

        mov     [esi+edx+TScroller.pos], eax
        exec    esi, TScrollable:__DrawScrollbar, edx
        exec    esi, TScrollable:__Invalidate, edx

.finish:
        popad
        return
endp

;_________________________________________________________________________________________


method TScrollable.SetPage         ; .param, .value
begin
        pushad

        mov     esi, [.self]

        mov     edx, TScrollable._scrX
        mov     ecx, TScrollable._scrY
        cmp     [.param], TScrollable.PageX
        cmovne  edx, ecx

        mov     eax, [.value]
        cmp     eax, [esi+edx+TScroller.page]
        je      .finish

        mov     [esi+edx+TScroller.page], eax
        exec    esi, TScrollable:__DrawScrollbar, edx
        exec    esi, TScrollable:__Invalidate, edx

.finish:
        popad
        return
endp

;_________________________________________________________________________________________


method TScrollable.SetMax          ;, .param, .value
begin
        pushad

        mov     esi, [.self]

        mov     edx, TScrollable._scrX
        mov     ecx, TScrollable._scrY
        cmp     [.param], TScrollable.MaxX
        cmovne  edx, ecx

        mov     eax, [.value]
        cmp     eax, [esi+edx+TScroller.max]
        je      .finish

        mov     [esi+edx+TScroller.max], eax
        exec    esi, TScrollable:__DrawScrollbar, edx
        exec    esi, TScrollable:__Invalidate, edx

.finish:
        popad
        return
endp



;_________________________________________________________________________________________


method TScrollable.SelfPaint       ;, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height
.src  RECT  ; the rectangle that need paint.
.vscr RECT  ; the vertical scroller rect.
.hscr RECT  ; the horizontal scroller rect.
.res  RECT
begin
        inherited [.pDstImage], [.xDst], [.yDst], [.xSrc], [.ySrc], [.width], [.height]
        pushad

        mov     esi, [.self]

        mov     eax, [.xSrc]
        mov     ecx, [.ySrc]
        mov     [.src.left], eax
        mov     [.src.top], ecx
        add     eax, [.width]
        add     ecx, [.height]
        mov     [.src.right], eax
        mov     [.src.bottom], ecx

        xor     eax, eax
        mov     [.vscr.top], eax
        mov     [.hscr.left], eax

        mov     ecx, [esi+TScrollable._width]
        mov     edx, [esi+TScrollable._height]

        mov     [.vscr.right], ecx
        mov     [.vscr.bottom], edx

        mov     [.hscr.right], ecx
        mov     [.hscr.bottom], edx

        sub     ecx, [GUI.scrollWidth]
        sub     edx, [GUI.scrollWidth]

        mov     [.vscr.left], ecx
        mov     [.hscr.top], edx

        lea     edx, [.src]
        lea     ebx, [.vscr]
        lea     eax, [.res]

        stdcall RectIntersect, eax, ebx, edx
        jc      .vscr_ok

        cmp     [esi+TScrollable._scrY.max], 0
        je      .vscr_ok

        mov     eax, [.res.left]
        mov     ebx, [.res.top]
        mov     ecx, eax
        mov     edx, ebx
        sub     [.res.right], eax
        sub     [.res.bottom], ebx
        sub     eax, [.vscr.left]
        sub     ebx, [.vscr.top]

        add     ecx, [.xDst]
        add     edx, [.yDst]
        sub     ecx, [.xSrc]
        sub     edx, [.ySrc]

        stdcall BlendImage, [.pDstImage], ecx, edx, [esi+TScrollable._imgY], eax, ebx, [.res.right], [.res.bottom]

.vscr_ok:
        lea     edx, [.src]
        lea     ebx, [.hscr]
        lea     eax, [.res]
        stdcall RectIntersect, eax, ebx, edx
        jc      .hscr_ok

        cmp     [esi+TScrollable._scrX.max], 0
        je      .hscr_ok

        mov     eax, [.res.left]
        mov     ebx, [.res.top]
        mov     ecx, eax
        mov     edx, ebx
        sub     [.res.right], eax
        sub     [.res.bottom], ebx
        sub     eax, [.hscr.left]
        sub     ebx, [.hscr.top]

        add     ecx, [.xDst]
        add     edx, [.yDst]
        sub     ecx, [.xSrc]
        sub     edx, [.ySrc]

        stdcall BlendImage, [.pDstImage], ecx, edx, [esi+TScrollable._imgX], eax, ebx, [.res.right], [.res.bottom]

.hscr_ok:
        popad
        return
endp



method TScrollable.__DrawScrollbar
begin
        pushad
        mov     esi, [.self]

        cmp     [.scrollbar], TScrollable._scrX
        je      .draw_x

        mov     ebx, [esi+TScrollable._imgY]
        test    ebx, ebx
        jz      .scry_ok

        mov     edi, [esi+TScrollable._scrY.state]

; Draw background Y

        mov     ecx, [ebx+TImage.wrapH]
        mov     edx, [GUI.scrollWidth]
        sub     ecx, edx
        stdcall DrawSolidRect, ebx, 0, 0, edx, ecx, [GUI.clScrollBk+4*edi]
        stdcall DrawSolidRect, ebx, 0, ecx, edx, edx, 0
        stdcall BlendTriangle, ebx, 0, ecx, edx, trtUpper or trtRight, [GUI.clScrollBk+4*edi]

; Draw slider Y
        lea     eax, [esi+TScrollable._scrY]
        stdcall _SliderPixels2, eax, ecx

        cmp     eax, ecx
        cmovg   eax, ecx

        mov     ecx, [esi+TScrollable._scrY.pos]
        cmp     ecx, [esi+TScrollable._scrY.max]
        cmovg   ecx, [GUI.clScrollSliderOvf+4*edi]
        cmovle  ecx, [GUI.clScrollSlider+4*edi]

        stdcall BlendSolidRect, ebx, 0, eax, [GUI.scrollWidth], edx, ecx
        add     eax, edx
        stdcall BlendTriangle,  ebx, 0, eax, [GUI.scrollWidth], trtUpper or trtRight, ecx

.scry_ok:
        popad
        return


.draw_x:
        mov     ebx, [esi+TScrollable._imgX]
        test    ebx, ebx
        jz      .scrx_ok

        mov     edi, [esi+TScrollable._scrX.state]

; Draw background X

        mov     ecx, [ebx+TImage.wrapW]
        mov     edx, [GUI.scrollWidth]
        sub     ecx, edx
        stdcall DrawSolidRect, ebx, 0, 0, ecx, edx, [GUI.clScrollBk+4*edi]
        stdcall DrawSolidRect, ebx, ecx, 0, edx, edx, 0
        stdcall BlendTriangle, ebx, ecx, 0, edx, trtLower or trtLeft, [GUI.clScrollBk+4*edi]

; Draw slider X

        lea     eax, [esi+TScrollable._scrX]
        stdcall _SliderPixels2, eax, ecx

        cmp     eax, ecx
        cmovg   eax, ecx

        mov     ecx, [esi+TScrollable._scrX.pos]
        cmp     ecx, [esi+TScrollable._scrX.max]
        cmovg   ecx, [GUI.clScrollSliderOvf+4*edi]
        cmovle  ecx, [GUI.clScrollSlider+4*edi]

        stdcall BlendSolidRect, ebx, eax, 0, edx, [GUI.scrollWidth], ecx
        add     eax, edx
        stdcall BlendTriangle,  ebx, eax, 0, [GUI.scrollWidth], trtLower or trtLeft, ecx

.scrx_ok:
        popad
        return
endp



; returns:
;   EAX:
;       -1 - the coordinates are outside the scrollbars
;        0 - the coordinates are in the horizontal scrollbar
;        1 - the coordinates are in the vertical scrollbar

proc __FindScroller, .scrollable, .x, .y
begin
        push    esi ecx edx

        mov     esi, [.scrollable]

        xor     eax, eax
        dec     eax

        mov     ecx, [esi+TWindow._width]
        mov     edx, [esi+TWindow._height]
        sub     ecx, [.x]
        sub     edx, [.y]

        cmp     ecx, [GUI.scrollWidth]
        jl      .y_scroller

        cmp     edx, [GUI.scrollWidth]
        jge     .finish

        cmp     ecx, edx
        jl      .yes_y_scroller

.yes_x_scroller:

        inc     eax
        jmp     .finish

.y_scroller:

        cmp     ecx, edx
        jge     .yes_x_scroller

.yes_y_scroller:

        inc     eax
        inc     eax

.finish:
        pop     edx ecx esi
        return
endp




method TScrollable.EventMouseLeave
begin
        push    esi ecx edx

        mov     esi, [.self]

        xor     ecx, ecx
        xor     edx, edx
        xchg    ecx, [esi+TScrollable._scrX.state]
        xchg    edx, [esi+TScrollable._scrY.state]
        or      ecx, edx
        jz      .finish

        exec    esi, TScrollable:__DrawScrollbar, TScrollable._scrX
        exec    esi, TScrollable:__DrawScrollbar, TScrollable._scrY
        exec    esi, TScrollable:__Invalidate, TScrollable._scrX
        exec    esi, TScrollable:__Invalidate, TScrollable._scrY

.finish:
        pop     edx ecx esi
        return
endp



method TScrollable.EventMouseMove
.dummy  dd ?
.stateH dd ?
.stateV dd ?
begin
        pushad

        mov     esi, [.self]
        mov     ebx, [esi+TScrollable._state]
        test    ebx, ebx
        jnz     .process_drag

        xor     ecx, ecx
        mov     [.stateH], ecx
        mov     [.stateV], ecx

        stdcall __FindScroller, esi, [.x], [.y]
        inc     [.stateH+4*eax]         ; notice the .dummy variable for eax = -1

        mov     ecx, [esi+TWindow._cursor]
        inc     eax
        jz      @f
        mov     ecx, mcArrow
@@:
        get     edx, esi, TWindow:OSWindow
        stdcall GetStockCursor, ecx
        stdcall SetMouseCursor, [edx+TWindow.handle], eax

        xor     eax, eax
        mov     ebx, [.stateH]
        xchg    ebx, [esi+TScrollable._scrX.state]
        xor     ebx, [esi+TScrollable._scrX.state]
        jz      .statex_ok

        exec    esi, TScrollable:__DrawScrollbar, TScrollable._scrX
        exec    esi, TScrollable:__Invalidate, TScrollable._scrX

.statex_ok:
        mov     ebx, [.stateV]
        xchg    ebx, [esi+TScrollable._scrY.state]
        xor     ebx, [esi+TScrollable._scrY.state]
        or      eax, ebx
        jz      .statey_ok   ; if the state has not be changed.

        exec    esi, TScrollable:__DrawScrollbar, TScrollable._scrY
        exec    esi, TScrollable:__Invalidate, TScrollable._scrY

.statey_ok:
        inherited  [.x], [.y], [.kbdState]
        clc
        popad
        return


.process_drag:
        dec     ebx
        lea     edi, [4*ebx]

        mov     eax, [.x + edi]
        mov     ecx, [esi+TScrollable._width + edi]

        sub     eax, [esi+TScrollable._drag_ofs]
        lea     edx, [esi+TScrollable._scrX + 4*edi]

        stdcall _ComputeSliderPos, edx, eax, ecx
        je      .finish


        mov     ecx, TScrollable.PosX
        test    edi, edi
        jz      .setpos
        mov     ecx, TScrollable.PosY
.setpos:
        exec    esi, TScrollable:SetPos, ecx, eax
        exec    esi, TScrollable:EventScroll, ebx, scTrack, [edx+TScroller.pos]

.finish:
        stc
        popad
        return
endp



; computes the position of the slider from screen pixel coordinates.
proc _ComputeSliderPos, .pScroller, .x, .lengthpx
begin
        push    ecx edx

        mov     ecx, [.pScroller]
        stdcall _SliderPixels2, ecx, [.lengthpx]

        sub     [.lengthpx], edx
        jz      .zero

        mov     eax, [ecx+TScroller.max]
        imul    [.x]
        idiv    [.lengthpx]

        xor     edx, edx

        cmp     eax, edx
        cmovl   eax, edx

        mov     edx, [ecx+TScroller.max]
        cmp     eax, edx
        cmovg   eax, edx

.ok:
        pop     edx ecx
        return

.zero:
        xor     eax, eax
        jmp     .ok
endp




method TScrollable.EventButtonPress
begin
        pushad

        cmp     [.button], mbLeft
        jne     .not_processed

        mov     esi, [.self]

        stdcall __FindScroller, esi, [.x], [.y]
        test    eax, eax
        js      .not_processed

        mov     ebx, eax
        shl     eax, 4
        lea     eax, [esi+TScrollable._scrX + eax]
        stdcall _SliderPixels2, eax, [esi+TScrollable._width + 4*ebx]
        mov     edi, [.x + 4*ebx]

        mov     ecx, [esi+TScrollable._width + 4*ebx]   ; if pos is > max
        sub     ecx, edx
        cmp     eax, ecx
        cmovg   eax, ecx

        sub     edi, eax
        jl      .before

        cmp     edi, edx
        jg      .after

; inside the slider - capture the mouse.

        mov     [esi+TScrollable._drag_ofs], edi
        inc     ebx
        mov     [esi+TScrollable._state], ebx

        stdcall SetMouseCapture, esi

.finish:
        inherited [.button], [.kbdState], [.x], [.y]

        stc
        popad
        return

.before:
.after:
        jmp     .finish


.not_processed:
        clc
        popad
        return
endp




method TScrollable.EventButtonRelease
begin
        pushad

        mov     esi, [.self]
        xor     eax, eax
        cmp     [esi+TScrollable._state], eax
        je      .finish

        mov     [esi+TScrollable._drag_ofs], eax
        mov     [esi+TScrollable._state], eax
        stdcall SetMouseCapture, eax

        exec    esi, TScrollable:EventMouseMove, [.x], [.y], [.kbdState]

.finish:
        popad
        return
endp



method TScrollable.EventScroll
begin
        pushad

        mov     esi, [.self]

        cmp     [.direction], scrollX
        je      .horiz_scroll

        cmp     [.command], scWheelUp
        je      .scrollupy

        cmp     [.command], scWheelDn
        je      .scrolldny

        cmp     [.command], scTrack
        jne     .finish

        set     esi, TScrollable:PosY, [.value]

.finish:
        popad
        return

.scrollupy:
        get     eax, esi, TScrollable:PosY
        dec     eax
        jns     @f
        xor     eax, eax
@@:
        set     esi, TScrollable:PosY, eax
        jmp     .finish


.scrolldny:
        get     ecx, esi, TScrollable:MaxY
        get     eax, esi, TScrollable:PosY

        inc     eax
        cmp     eax, ecx
        jle     @f
        dec     eax
@@:
        set     [.self], TScrollable:PosY, eax
        jmp     .finish


.horiz_scroll:
        cmp     [.command], scWheelUp
        je      .scrollupx

        cmp     [.command], scWheelDn
        je      .scrolldnx

        cmp     [.command], scTrack
        jne     .finish

        set     esi, TScrollable:PosX, [.value]
        jmp     .finish


.scrollupx:
        get     eax, esi, TScrollable:PosX
        dec     eax
        jns     @f
        xor     eax, eax
@@:
        set     esi, TScrollable:PosX, eax
        jmp     .finish


.scrolldnx:
        get     eax, esi, TScrollable:PosX
        get     ecx, esi, TScrollable:MaxX
        cmp     eax, ecx
        jg      @f
        inc     eax
@@:
        set     [.self], TScrollable:PosX, eax
        jmp     .finish
endp






; returns:
;   eax - position in pixels
;   edx - size in pixels


proc _SliderPixels2, .pScroller, .length
begin
        push    ecx esi

        mov     ecx, [.pScroller]

; page size in pixels.
        mov     esi, [ecx+TScroller.max]
        mov     eax, [.length]

        add     esi, [ecx+TScroller.page]
        test    esi, esi
        jz      .lengthok

        imul    [ecx+TScroller.page]
        idiv    esi

.lengthok:
        cmp     eax, [GUI.minSliderHeight]
        cmovl   eax, [GUI.minSliderHeight]

        cmp     eax, [.length]
        cmovg   eax, [.length]

        push    eax

; position in pixels

        sub     eax, [.length]
        mov     esi, [ecx+TScroller.max]
        test    esi, esi
        jnz     @f
        inc     esi
@@:
        neg     eax
        mul     [ecx+TScroller.pos]
        div     esi

        pop     edx   ; width

        pop     esi ecx
        return
endp









endmodule



