; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TToolbar object class
;
;  Target OS: Any
;
;  Dependencies:
;_________________________________________________________________________________________
module "TToolbar library"

tbbSeparator = 0

tbsNeutral  = bsNeutral
tbsHovered  = bsHovered
tbsPressed  = bsPressed
tbsDisabled = bsDisabled


struct TToolButton
  .action dd ?          ; can be pointer to TAction object, or TMenu or tbbSeparator (0)
  .state  dd ?
  .width  dd ?
ends


object TToolbar, TWindow

; private fields

  ._canvas       dd ?
  ._canvas_valid dd ?

  ._btn_height   dd ?

  ._buttons     dd ?            ; TArray of TToolButton structures.

; methods

  method .Create, .Parent
  method .ForceRefresh

  method .HitTest, .x, .y
  method .ClearState

  method .AddButton, .action

  method ._RenderView

  method .SelfPaint, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height

  method .Autosize

  ; system events methods

  method .EventMouseEnter
  method .EventMouseLeave
  method .EventMouseMove, .x, .y, .kbdState

  method .EventButtonPress, .button, .kbdState, .x, .y
  method .EventButtonRelease, .button, .kbdState, .x, .y

endobj


interface TToolbar.OnClick, .self, .button


; new

method TToolbar.Create
begin
        inherited [.Parent]
        push    esi

        mov     esi, [.self]
        mov     [esi+TToolbar._cursor], mcArrow
        mov     [esi+TToolbar.__want_focus], FALSE

        stdcall CreateArray, sizeof.TToolButton
        mov     [esi+TToolbar._buttons], eax
        mov     [esi+TToolbar._btn_height], 24

        pop     esi
        return
endp



method TToolbar.ForceRefresh
begin
        mov     eax, [.self]
        and     [eax+TToolbar._canvas_valid], 0
        inherited
        return
endp


method TToolbar.SelfPaint
begin
        pushad
        mov     esi, [.self]

        cmp     [esi+TToolbar._canvas_valid], 0
        jne     @f
        exec    esi, TToolbar:_RenderView
@@:
        mov     ebx, [esi+TToolbar._canvas]
        test    ebx, ebx
        jz      .finish

        stdcall DrawSolidRect, [.pDstImage], [.xDst], [.yDst], [.width], [.height], [GUI.clDialogBk]
        stdcall BlendImage, [.pDstImage], [.xDst], [.yDst], ebx, [.xSrc], [.ySrc], [.width], [.height]

.finish:
        popad
        inherited [.pDstImage], [.xDst], [.yDst], [.xSrc], [.ySrc], [.width], [.height]
        return
endp


method TToolbar.HitTest
begin
        pushad

        mov     esi, [.self]
        mov     edi, [esi+TToolbar._buttons]
        mov     ecx, [edi+TArray.count]
        lea     edi, [edi+TArray.array]

.loop:
        dec     ecx
        js      .not_found

        mov     eax, [edi+TToolButton.width]
        cmp     [.x], eax
        jl      .found

        sub     [.x], eax
        add     edi, sizeof.TToolButton
        jmp     .loop

.found:
        cmp     [edi+TToolButton.action], 0
        je      .not_found

        mov     [esp+4*regEAX], edi
        clc
        popad
        return

.not_found:
        stc
        popad
        return
endp


method TToolbar.AddButton
begin
        pushad

        mov     esi, [.self]
        mov     edx, [esi+TToolbar._buttons]
        mov     ecx, [.action]

        stdcall AddArrayItems, edx, 1
        mov     [esi+TToolbar._buttons], edx

        test    ecx, ecx
        cmovz   edx, [GUI.tbSeparatorWidth]
        cmovnz  edx, [esi+TToolbar._btn_height]

        mov     [eax+TToolButton.action], ecx
        mov     [eax+TToolButton.state], tbsNeutral
        mov     [eax+TToolButton.width], edx

        exec    esi, TToolbar:ForceRefresh
        popad
        return
endp



method TToolbar.ClearState
begin
        pushad
        mov     esi, [.self]
        mov     edi, [esi+TToolbar._buttons]
        mov     ecx, [edi+TArray.count]
        lea     edi, [edi+TArray.array]

.loop:
        dec     ecx
        js      .finish

        mov     [edi+TToolButton.state], tbsNeutral
        add     edi, sizeof.TToolButton
        jmp     .loop

.finish:
        popad
        return
endp



method TToolbar._RenderView
begin
        pushad

        mov     esi, [.self]
        mov     edi, [esi+TToolbar._buttons]
        mov     ecx, [edi+TArray.count]
        jecxz   .finish

        mov     ebx, [esi+TToolbar._canvas]
        test    ebx, ebx
        jnz     .canvas_ok

        mov     eax, ecx
        imul    eax, [esi+TToolbar._btn_height]
        stdcall CreateImage, eax, [esi+TToolbar._btn_height]
        mov     [esi+TToolbar._canvas], eax
        mov     ebx, eax

.canvas_ok:

        lea     edi, [edi + TArray.array]
        xor     edx, edx

.loop:
        dec     ecx
        js      .finish

        stdcall DrawToolbarBtn, ebx, edi, edx, 0, [esi+TToolbar._btn_height]

        add     edx, [edi+TToolButton.width]
        add     edi, sizeof.TToolButton
        jmp     .loop


.finish:
        inc     [esi+TToolbar._canvas_valid]
        popad
        return
endp




proc DrawToolbarBtn, .canvas, .pButton, .x, .y, .height
.bounds TBounds
begin
        pushad
        mov     edi, [.pButton]

        mov     eax, [.x]
        mov     ecx, [.y]
        mov     [.bounds.x], eax
        mov     [.bounds.y], ecx

        mov     eax, [edi+TToolButton.width]
        mov     ecx, [.height]
        mov     [.bounds.width], eax
        mov     [.bounds.height], ecx

        mov     ebx, [edi+TToolButton.state]

        lea     eax, [.bounds]
        stdcall [DrawBox], [.canvas], eax, [GUI.clToolBtnBk+4*ebx], [GUI.tbBorder+4*ebx], [GUI.tbBorderWidth]

        cmp     [edi+TToolButton.action], 0
        je      .draw_separator

        istype  [edi+TToolButton.action], TAction
        jne     .finish

        add     [.bounds.x], 4
        add     [.bounds.y], 4

        mov     eax, [GUI.tbPressedOfsX]
        mov     edx, [GUI.tbPressedOfsY]

        cmp     [edi+TToolButton.state], tbsHovered
        jne     @f
        neg     eax
        neg     edx
@@:
        cmp     [edi+TToolButton.state], tbsPressed
        je      @f
        cmp     [edi+TToolButton.state], tbsHovered
        je      @f
        xor     eax, eax
        xor     edx, edx
@@:
        add     [.bounds.x], eax
        add     [.bounds.y], edx

        get     ecx, [edi+TToolButton.action], TAction:Icon
        test    ecx, ecx
        jz      .finish

        stdcall BlendImage, [.canvas], [.bounds.x], [.bounds.y], ecx, 0, 0, [ecx+TImage.width], [ecx+TImage.height]

.finish:
        popad
        return

.draw_separator:

        mov     ecx, [edi+TToolButton.width]
        sub     ecx, 2
        sub     [.bounds.height], ecx
        shr     ecx, 1
        add     [.bounds.x], ecx
        add     [.bounds.y], ecx
        mov     [.bounds.width], 2

        lea     eax, [.bounds]
        stdcall DrawBoxDefault, [.canvas], eax, 0, [GUI.tbSeparatorType], 1
;        stdcall DrawSolidRect, [.canvas], ecx, [.bounds.y], 2, [.bounds.height], [GUI.clToolSeparator]
;        stdcall DrawSolidRect, [.canvas], [.bounds.x], 0, [edi+TToolButton.width], 2, clAlizarin

        jmp     .finish


endp



method TToolbar.Autosize
begin
        inherited
        mov     edx, [.self]
        mov     edx, [edx+TToolbar._btn_height]
        return
endp



method TToolbar.EventMouseEnter
begin
;        exec    [.self], TToolbar:SimpleSetParam, 0, TToolbar._state, bsHovered
        inherited
        return
endp


method TToolbar.EventMouseLeave
begin
        exec    [.self], TToolbar:ClearState
        exec    [.self], TToolbar:ForceRefresh
        inherited
        return
endp


method TToolbar.EventMouseMove
begin
        test    [.kbdState], maskBtnLeft
        jnz     .exit

        exec    [.self], TToolbar:ClearState
        exec    [.self], TToolbar:HitTest, [.x], [.y]
        jc      .finish

        mov     [eax+TToolButton.state], bsHovered

.finish:
        exec    [.self], TToolbar:ForceRefresh

.exit:
        inherited [.x], [.y], [.kbdState]
        return
endp


method TToolbar.EventButtonPress
begin
        exec    [.self], TToolbar:ClearState
        exec    [.self], TToolbar:HitTest, [.x], [.y]
        jc      .finish

        mov     [eax+TToolButton.state], bsPressed
        stdcall SetMouseCapture, [.self]

.finish:
        exec    [.self], TToolbar:ForceRefresh
        return
endp



method TToolbar.EventButtonRelease
begin
        pushad
        stdcall SetMouseCapture, 0

        mov     esi, [.self]
        mov     edi, [esi+TToolbar._buttons]
        mov     ecx, [edi+TArray.count]
        lea     edi, [edi+TArray.array]

.loop:
        dec     ecx
        js      .finish

        cmp     [edi+TToolButton.state], tbsPressed
        je      .end_scan

        add     edi, sizeof.TToolButton
        jmp     .loop

.end_scan:
        exec    esi, TToolbar:HitTest, [.x], [.y]
        jc      .finish

        cmp     eax, edi
        jne     .finish

        mov     [edi+TToolButton.state], bsHovered
        exec    esi, TToolbar:ForceRefresh

        istype  [edi+TToolButton.action], TAction
        jne     .finish

        exec    [edi+TToolButton.action], TAction:Execute, esi

.finish:
        popad
        return
endp



endmodule