; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: This file contains procedures and data, that form the appearance of the
;               TAsmEdit components.
;
;  Target OS: Any
;
;  Dependencies:
;_________________________________________________________________________________________

module "FASM syntax highlighter for TAsmEdit"


ciText         = 0
ciComment      = 1
ciString       = 2
ciNumber       = 3
ciSymbol       = 4
ciRegister     = 5
ciInstruction  = 6
ciDerective    = 7
ciPreprocessor = 8
ciMacro        = 9
ciWhiteSpace   = 1


;_________________________________________________________________________________________
;
; Arguments:
;   .pLineInfo - pointer to TTextLine structure.
;   .SynContext - input context of the highligher.
;
; Returns:
;       eax - output context of the highligher.
;
proc SyntaxFASM, .pTextLine, .SynContext
.Quo   dd ?
.cnt   dd ?
begin
        pushad

        mov     esi, [.pTextLine]

        mov     edi, [esi+TTextLine.syn_array]
        test    edi, edi
        jnz     .array_ok

        stdcall CreateArray, sizeof.TTextChunk
        mov     edi, eax
        mov     [esi+TTextLine.syn_array], eax

.array_ok:

        xor     ebx, ebx                ; the offset inside the line.
        xor     edx, edx

        mov     [.Quo], ebx
        mov     [edi+TArray.count], ebx
        mov     [.cnt], ebx
        dec     [.cnt]

.loop:
        add     ebx, edx
        inc     [.cnt]

        mov     eax, [esi+TTextLine.GapBegin]
        cmp     ebx, eax
        cmovae  eax, [esi+TTextLine.GapEnd]
        sub     eax, [esi+TTextLine.GapBegin]
        add     eax, ebx

        cmp     eax, [esi+TTextLine.Length]
        jae     .end_of_line

        stdcall DecodeUtf8, [esi+eax]

;        OutputValue "Char: ", eax, 16, 8

        cmp     [.SynContext], ciComment
        je      .loop

        cmp     [.SynContext], ciString
        jne     .ifnotstr

        cmp     eax,[.Quo]
        jne     .loop

        add     ebx, edx
        xor     edx, edx
        inc     [.cnt]
        call    .end_chunk
        mov     [.SynContext], ciText
        jmp     .loop

.ifnotstr:
        cmp     eax, $ff
        ja      .text

        cmp     eax, ';'
        jne     .notrem

        call    .end_chunk
        mov     [.SynContext], ciComment
        jmp     .loop

.notrem:
        cmp     eax, '"'
        je      .str
        cmp     eax, "'"
        jne     .notstr

.str:
        call    .end_chunk
        mov     [.SynContext], ciString
        mov     [.Quo], eax
        jmp     .loop

.notstr:
        call    .CharInSymbols
        jnc     .notsym

        cmp     [.SynContext], ciSymbol
        je      .loop

        call    .end_chunk
        mov     [.SynContext], ciSymbol
        jmp     .loop

.notsym:
        cmp     [.SynContext], ciNumber
        je      .loop

        cmp     [.cnt], 0
        je      .maybenum

        cmp     [.SynContext], ciSymbol
        jne     .text

; The number begins only after ttChar or at the begining of the line.
.maybenum:
        cmp     al, '$'
        je      .num_start

        cmp     al, '-'
        je      .num_start

        cmp     al, '0'
        jb      .text
        cmp     al, '9'
        ja      .text

.num_start:
        call    .end_chunk
        mov     [.SynContext], ciNumber
        jmp     .loop

.text:
        cmp     [.SynContext], ciText
        je      .loop

        call    .end_chunk
        mov     [.SynContext], ciText
        jmp     .loop


.end_of_line:
        call    .end_chunk

;        mov     eax, [.SynContext]
        xor     eax, eax                ; the assembly lines always end with context of 0 ???

        mov     [esi+TTextLine.syn_array], edi
        mov     [esi+TTextLine.syn_context], eax
        mov     [esp+4*regEAX], eax
        popad
        return


.end_chunk:
        push    eax ecx edx

        mov     ecx, [.cnt]
        test    ecx, ecx
        jz      .chunk_processed

        and     [.cnt], 0

        stdcall AddArrayItems, edi, 1
        mov     edi, edx
        jc      .chunk_processed

;        OutputValue "Chunk length: ", ecx, 16, 8

        mov     [eax+TTextChunk.len], cx
        mov     ecx, [.SynContext]
        mov     [eax+TTextChunk.color], cl

;        OutputValue "Chunk color: ", ecx, 16, 8

.chunk_processed:
        pop     edx ecx eax
        retn


.CharInSymbols:
        push    edi ecx

        mov     edi, .symbol_characters+1
        movzx   ecx, byte [edi-1]

        repne scasb
        jne   @f
        stc
        pop     ecx edi
        ret
@@:
        clc
        pop     ecx edi
        retn

.symbol_characters db 27
                  db 9,0Ah,0Dh,1Ah,20h,'+-/*=<>()[]{}:,|&~#`;\'

endp




;proc SyntaxFASM2, .pTextLine, .SynContext
;.Quo     rb 4
;.start   dd ?
;.current TTextChunk
;begin
;        pushad
;
;        mov     esi, [.pTextLine]
;
;        mov     edi, [esi+TTextLine.syn_array]
;        test    edi, edi
;        jnz     .array_ok
;
;        stdcall CreateArray, sizeof.TTextChunk
;        mov     edi, eax
;        mov     [esi+TTextLine.syn_array], eax
;
;.array_ok:
;
;        xor     ebx, ebx                ; the offset inside the line.
;        xor     edx, edx
;
;        mov     dword [.current], ebx
;        mov     dword [.current+4], ebx
;
;        mov     [.Quo], bl
;        mov     [edi+TArray.count], ebx
;        mov     [.start], ebx
;
;.loop:
;        add     ebx, edx
;
;        mov     eax, ebx
;        pos2ofs eax, esi
;
;        cmp     eax, [esi+TTextLine.Length]
;        jae     .end_of_line
;
;        stdcall DecodeUtf8, [esi+eax]
;
;        cmp     [.SynContext], ciComment
;        je      .endif
;
;        cmp     [.SynContext], ciString
;        jne     .ifnotstr
;
;        cmp     al,[.Quo]
;        jne     .endif
;
;        add     ebx, edx
;        xor     edx, edx
;        call    .end_chunk
;        mov     [.SynContext], ciText
;        jmp     .loop
;
;.ifnotstr:
;        cmp     al, ';'
;        jne     .notrem
;
;        call    .end_chunk
;        mov     [.SynContext], ciComment
;        jmp     .endif
;
;.notrem:
;        cmp     al, '"'
;        je      .str
;        cmp     al, "'"
;        jne     .notstr
;
;.str:
;        call    .end_chunk
;        mov     [.SynContext], ciString
;        mov     [.Quo], al
;        jmp     .endif
;
;.notstr:
;        call    .CharInSymbols
;        jnc     .notsym
;
;        call    .end_chunk
;        mov     [.SynContext], ciSymbol
;        jmp     .endif
;
;.notsym:
;        cmp     [.SynContext], ciNumber
;        je      .endif
;
;        cmp     ebx, [.start]
;        je      .maybenum
;
;        cmp     [.SynContext], ciSymbol
;        jne     .endif
;
;; The number begins only after ttChar or at the begining of the line.
;.maybenum:
;        call    .CharInNum
;        jnc     .notnum
;
;        call    .end_chunk
;        mov     [.SynContext], ciNumber
;        jmp     .endif
;
;.notnum:
;        call    .end_chunk
;        mov     [.SynContext], ciText
;
;.endif:
;        jmp     .loop
;
;
;.end_of_line:
;        call    .end_chunk
;
;;        mov     eax, [.SynContext]
;        xor     eax, eax                ; the assembly lines always end with context of 0 ???
;
;        mov     [esi+TTextLine.syn_array], edi
;        mov     [esi+TTextLine.syn_context], eax
;        mov     [esp+4*regEAX], eax
;        popad
;        return
;
;; ecx = symbol chunk type.
;; edx = length of the character in bytes.
;
;.add_to_chunk:
;        cmp     [.current.color], cl
;        jne     .add_to_list            ; add the current to the list and start new chunk
;
;        add     [.current.len], dx
;        retn
;
;.add_to_list:
;        push    edx
;        stdcall AddArrayItems,edi, 1
;        mov     edi, edx
;
;        mov     edx, [.current]
;        mov     [eax], edx
;        mov     edx, [.current+4]
;        mov     [eax+4], edx
;
;        pop     edx
;
;        mov     [.current.len], dx
;        mov     [.current.color], cl
;        mov     [.current.flags], 0
;        retn
;
;
;.end_chunk:
;        push    eax ecx edx
;
;        mov     ecx, ebx
;        sub     ecx, [.start]
;        jz      .chunk_processed
;
;        mov     [.start], ebx
;        stdcall AddArrayItems, edi, 1
;        mov     edi, edx
;        jc      .chunk_processed
;
;        mov     [eax+TTextChunk.len], cx
;        mov     ecx, [.SynContext]
;        mov     [eax+TTextChunk.color], cl
;
;.chunk_processed:
;        pop     edx ecx eax
;        retn
;
;
;.CharInSymbols:
;        push    edi ecx
;
;        mov     edi, .symbol_characters+1
;        movzx   ecx, byte [edi-1]
;
;        repne scasb
;        jne   @f
;        stc
;        pop     ecx edi
;        ret
;@@:
;        clc
;        pop     ecx edi
;        retn
;
;
;.CharInNum:
;        cmp     al, '$'
;        jne     @f
;        stc
;        retn
;
;@@:
;        cmp     al, '0'
;        jb      @f
;        cmp     al, '9'
;        ja      @f
;        stc
;        retn
;@@:
;        clc
;        retn
;
;
;.symbol_characters db 27
;                  db 9,0Ah,0Dh,1Ah,20h,'+-/*=<>()[]{}:,|&~#`;\'
;
;endp











endmodule