; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Defines constants and structures for system events of FreshLib
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________
module "System events library"


mbLeft          = 1
mbMiddle        = 2
mbRight         = 3

mbScrollUp      = 4
mbScrollDn      = 5
mbScrollLeft    = 6
mbScrollRight   = 7


maskBtnLeft   = $0100
maskBtnMiddle = $0200
maskBtnRight  = $0400
maskShift     = $1
maskCapsLock  = $2
maskCtrl      = $4
maskAlt       = $8
maskScrLk     = $10


; These are scan codes for the non character keys. They are OS dependent and should be used by
; their names.
include '%TargetOS%/keycodes.inc'


scTrack         = 1
scUp            = 2
scDown          = 3
scWheelUp       = 4
scWheelDn       = 5

scrollX = 0
scrollY = 1

cerFromUser = 0         ; user pressed X button on the window header, pressed alt+F4 from the keyboard, or choose "Close" from the title menu.
cerFromProgram = 1      ; Method Close was executed from the program.



; creates and returns a string with the textual identifier of the pressed/released key.

proc CreateKeyName, .keycode, .kbdStatus
begin
        push    ebx

        stdcall StrNew
        mov     ebx, eax

        test    [.kbdStatus], maskCtrl
        jz      @f
        stdcall StrCat, ebx, txt 'Ctrl+'
@@:
        test    [.kbdStatus], maskAlt
        jz      @f
        stdcall StrCat, ebx, txt 'Alt+'
@@:
        test    [.kbdStatus], maskShift
        jz      @f
        stdcall StrCat, ebx, txt 'Shift+'
@@:
        stdcall KbdCodeToName, [.keycode]
        jc      .finish

        stdcall StrCat, ebx, eax
        stdcall StrDel, eax
        mov     eax, ebx
        clc

.finish:
        pop     ebx
        return
endp


macro keynames lbl, [scancode, name] {
common
  label lbl word
forward
  local ..pname
        dw scancode
        dw ..pname - $ + 2
common
        dw 0
forward
  ..pname db name, 0
}


iglobal
  if used __FunctionalKeyNames
  keynames  __FunctionalKeyNames,                        \
            keyHomeNumpad,     'Home',                   \
            keyUpNumpad,       'Up',                     \
            keyPgUpNumpad,     'PgUp',                   \
            keyLeftNumpad,     'Left',                   \
            key5Numpad,        'Num5',                   \
            keyRightNumpad,    'Right',                  \
            keyEndNumpad,      'End',                    \
            keyDownNumpad,     'Down',                   \
            keyPgDnNumpad,     'PgDn',                   \
            keyInsNumpad,      'Ins',                    \
            keyDelNumpad,      'Del',                    \
            keyEnterNumpad,    'Enter',                  \
            keyNumLock,        'NumLock',                \
            keyScrollLock,     'ScrLock',                \
            keyPause,          'Pause',                  \
            keyLeft,           'Left',                   \
            keyRight,          'Right',                  \
            keyUp,             'Up',                     \
            keyDown,           'Down',                   \
            keyInsert,         'Ins',                    \
            keyDelete,         'Del',                    \
            keyHome,           'Home',                   \
            keyEnd,            'End',                    \
            keyPgUp,           'PgUp',                   \
            keyPgDown,         'PgDn',                   \
            keyF1,             'F1',                     \
            keyF2,             'F2',                     \
            keyF3,             'F3',                     \
            keyF4,             'F4',                     \
            keyF5,             'F5',                     \
            keyF6,             'F6',                     \
            keyF7,             'F7',                     \
            keyF8,             'F8',                     \
            keyF9,             'F9',                     \
            keyF10,            'F10',                    \
            keyF11,            'F11',                    \
            keyF12,            'F12',                    \
            keyCapsLock,       'CapsLock',               \
            keyShiftLeft,      'ShiftL',                 \
            keyCtrlLeft,       'LeftCtrl',               \
            keyAltLeft,        'LeftAlt',                \
            keyAltRight,       'RightAlt',               \
            keyPopupMenu,      'PopupMenu',              \
            keyShiftRight,     'RightShift',             \
            keyCtrlRight,      'RightCtrl',              \
            keyBackSpace,      'BkSp',                   \
            keyTab,            'Tab',                    \
            keyA,              'A',                      \
            keyB,              'B',                      \
            keyC,              'C',                      \
            keyD,              'D',                      \
            keyE,              'E',                      \
            keyF,              'F',                      \
            keyG,              'G',                      \
            keyH,              'H',                      \
            keyI,              'I',                      \
            keyJ,              'J',                      \
            keyK,              'K',                      \
            keyL,              'L',                      \
            keyM,              'M',                      \
            keyN,              'N',                      \
            keyO,              'O',                      \
            keyP,              'P',                      \
            keyQ,              'Q',                      \
            keyR,              'R',                      \
            keyS,              'S',                      \
            keyT,              'T',                      \
            keyU,              'U',                      \
            keyV,              'V',                      \
            keyW,              'W',                      \
            keyX,              'X',                      \
            keyY,              'Y',                      \
            keyZ,              'Z'
  end if
endg

endmodule





proc KbdCodeToName, .keycode
begin
        pushad
        mov     esi, __FunctionalKeyNames

.loop:
        movzx   ecx, word [esi]
        jecxz   .notfound

        cmp     ecx, [.keycode]
        je      .found
        add     esi, 4
        jmp     .loop

.found:
        movsx   eax, word [esi+2]
        add     esi, eax
        stdcall StrDupMem, esi
        mov     [esp+4*regEAX], eax
        clc
        popad
        return

.notfound:
        stc
        popad
        return
endp




proc KbdNameToCode, .hKeyname
begin
        pushad
        mov     esi, __FunctionalKeyNames

.loop:
        movzx   ecx, word [esi]
        jecxz   .notfound

        movsx   eax, word [esi+2]
        lea     eax, [esi+eax]

        add     esi, 4
        stdcall StrCompNoCase, eax, [.hKeyname]
        jnc     .loop

.found:
        stdcall StrDupMem, esi
        mov     [esp+4*regEAX], ecx
        clc
        popad
        return

.notfound:
        stc
        popad
        return
endp
