; Pearsons hash function table:

struct TPHashItem
  .pKeyname    dd ?     ; pointer to the key name. on [.pKeyname-1] is 1 byte key length.
  .Value       dd ?     ; user defined value of the hash table cell.
ends

;
; Builds static Pearsons hash table.
; The table contains 256 elements of TPHashItem (in total 2048 bytes)
;
; Arguments:
;   table - the lable of the hash table.
;   Pfunc - table of 256 random (or not random) bytes - values of the Pearsons hash function.
;   key   - string that to be hashed. Pointer to this string is placed in the .pKeyname field of
;           the respective record of the hash table. The hash of the key is computed case insensitive.
;   value - user data. Placed in the .Value field of the respective TPHashItem element.
;
; Notice, that the table does not resolve hash collisions. If collision happens, the compilation ends
; with error. The conflict can be resolved in two ways - changing the hash function in order to
; avoid collisions, or changing the key that causes collision.
;
macro PHashTable table, Pfunc, [key, value] {
common
if used table
  local ..error
  ..error = 0

  if defined options.ShowHashes & options.ShowHashes
    disp 3, 'Hash table "', `table, '"', 9
  end if

  table dd 256 dup(0,0)

forward
local ..keynm, ..len, ..hash, ..char, ..prev

          db ..len
  ..keynm db key
  ..len = $ - ..keynm
         db 0

  if value
    ..hash = 0
    repeat ..len
      load ..char byte from ..keynm + % - 1
      if ..char and $40
        ..char = ..char or $20
      end if
      ..hash = ..hash xor ..char
      load ..hash byte from Pfunc + ..hash
      store byte ..char at ..keynm + % - 1
    end repeat

    if defined options.ShowHashes & options.ShowHashes
      disp 3,'Keyword hash : ', <..hash, 10>, ' on "', key, '"', 10
    end if

    load ..prev dword from table + ..hash * 8

    if ..prev = 0
      store dword ..keynm at table + ..hash * 8
      store dword value at table + ..hash * 8 + 4
    else
      disp 2,'Hash collision: ', <..hash, 10>, ' on "', key, '"', 10
      ..error = 1
    end if
  end if

common
  if defined options.ShowHashes & options.ShowHashes
    disp 6, '---', 13
  end if
  assert ~..error

  end if
}

;
; Computes the Pearsons hash in the same way as PHashTable macro.
; Useful for defining constants with the hash value of some key.
;
; usage:
;   MyHashConst phash PFUNK, "my_string"
;
struc phash Pfunc, key {
local ..keynm, ..len, ..hash, ..char

  virtual at 0
    ..keynm::
      db key
    ..len = $
  end virtual

  ..hash = 0
  repeat ..len
    load ..char byte from ..keynm:(% - 1)
    if ..char and $40
      ..char = ..char or $20
    end if
    ..hash = ..hash xor ..char
    load ..hash byte from Pfunc + ..hash
  end repeat

  if defined options.ShowHashes & options.ShowHashes
    disp 3,'Keyword hash : ', <..hash, 10>, ' on "', key, '"', 10
  end if

  . = ..hash
}
