; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Clipboard management library.
;
;  Target OS: Linux
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________


if used ClipboardRead | used ClipboardWrite
  __UseClipboard = 1
else
  __UseClipboard = 0
end if



body ClipboardRead
.result dd ?

.event XEvent

.return_type    dd ?
.return_format  dd ?
.return_items   dd ?
.return_remains dd ?
.return_data    dd ?

begin
        pushad
        mov     [.result], 0

        cinvoke XGetSelectionOwner, [hApplicationDisplay], [atomSelClipboard]
        test    eax, eax
        jz      .finish

        OutputValue "XGetSelectionOwner returns: ", eax, 10, -1

        cinvoke XConvertSelection, [hApplicationDisplay], [atomSelClipboard], [atomUTF8], [atomSelProperty], [hClipboardWindow], CurrentTime

        OutputValue "XConvertSelection returns: ", eax, 10, -1

        stdcall GetTimestamp
        mov     ebx, eax

.wait_loop:
        lea     eax, [.event]
        cinvoke XCheckTypedEvent, [hApplicationDisplay], SelectionNotify, eax
        test    eax, eax
        jnz     .notify_event

        stdcall ProcessSystemEvents

        stdcall GetTimestamp
        sub     eax, ebx
        cmp     eax, 1000
        jb      .wait_loop
        jmp     .timeout

.notify_event:

        DebugMsg "ClipboardRead SelectionNotify received!!!!!!!!"

        cmp     [.event+XSelectionEvent.property], None
        je      .finish

        lea     esi, [.return_data]
        lea     edx, [.return_remains]
        lea     ecx, [.return_items]
        lea     ebx, [.return_format]
        lea     eax, [.return_type]

        cinvoke XGetWindowProperty, [hApplicationDisplay], [.event+XSelectionEvent.requestor], [.event+XSelectionEvent.property], 0, $7fffffff, FALSE, [atomUTF8], eax, ebx, ecx, edx, esi
        OutputValue "XGetWindowProperty returns ", eax, 10, -1

        mov     eax, [.return_type]

        OutputValue "Return type: ", eax, 10, -1

        cmp     eax, [atomUTF8]
        jne     .finish

        mov      eax, [.return_data]

        OutputValue "Return data: ", eax, 10, -1

        test     eax, eax
        jz       .finish

        stdcall StrDup, eax
        mov     [.result], eax

        cinvoke XFree, [.return_data]

.finish:
        clc
        popad
        mov     eax, [.result]
        return

.timeout:
        stc
        popad
        return
endp



body ClipboardWrite   ; .hstring
begin
        pushad

        stdcall StrDup, [.hstring]
        mov     ecx, eax

        xchg    ecx, [Clipboard]
        stdcall StrDel, ecx     ; delete the previous if any.

        cinvoke XSetSelectionOwner, [hApplicationDisplay], [atomSelClipboard], [hClipboardWindow], CurrentTime
        OutputValue "XSetSelectionOwner returns: ", eax, 10, -1

        popad
        return
endp




uglobal
  var Clipboard = 0

  var atomSelClipboard = 0
  var atomUTF8 = 0
  var atomString = 0
  var atomSelProperty = 0
  var atomTargets = 0
  var atomAtom = 0

  var hClipboardWindow = 0
endg


proc __InitLinuxClipboard
begin
        cinvoke XInternAtom, [hApplicationDisplay], "CLIPBOARD", FALSE
        mov     [atomSelClipboard], eax

        cinvoke XInternAtom, [hApplicationDisplay], "TARGETS", FALSE
        mov     [atomTargets], eax

        cinvoke XInternAtom, [hApplicationDisplay], txt "ATOM", FALSE
        mov     [atomAtom], eax

        cinvoke XInternAtom, [hApplicationDisplay], "UTF8_STRING", FALSE
        mov     [atomUTF8], eax

        cinvoke XInternAtom, [hApplicationDisplay], "STRING", FALSE
        mov     [atomString], eax

        cinvoke XInternAtom, [hApplicationDisplay], "CUT_BUFFER0", FALSE
        mov     [atomSelProperty], eax

        cinvoke XCreateWindow, [hApplicationDisplay], [hRootWindow], 0, 0, 1, 1, 0, CopyFromParent, InputOnly, CopyFromParent, 0, 0
        mov     [hClipboardWindow], eax
        return
endp

