; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Process management library. Linux part
;
;  Target OS: Linux
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________


; This procedure switches the program in PER_LINUX32_3GB personality, because of
; some conflicts with StrLib handles that are placed in the number space above
; $c0000000.
;
; It is usually the first of the initialization procedures.

if defined options.Force3GB & options.Force3GB

proc ___SwitchLinuxTo3GB
begin
        OutputValue "Stack top address: $", esp, 16, 8

        cmp     esp, $c0000000
        jb      .finish                 ; the system is 32bit

        mov     eax, sys_personality
        mov     ebx, -1
        int     $80

        OutputValue "Initial personality: $", eax, 16, 8

        test    eax, ADDR_LIMIT_3GB
        jnz     .finish                         ; everything is OK.

; set the needed personality

        DebugMsg "Switch to 3GB address space."

        mov     eax, sys_personality
        mov     ebx, PER_LINUX32_3GB
        int     $80

        OutputValue "sys_personality return EAX:", eax, 10, -1

        test    eax, eax
        js      .finish         ; error set personality; try this way...???

; and restart the process

        mov     eax, [esp+4]          ; argument count
        mov     ebx, [esp+8]          ; the first argument is the name of the program.
        lea     ecx, [esp+8]          ; the arguments list.
        lea     edx, [ecx+4*eax+4]    ; the environment list.

        mov     eax, sys_execve
        int     $80

   ; !!! THIS PROC EXITS ONLY IF EVERYTHING IS OK !!!
   ; if something gone wrong, it comes here and stops!

        int3

.finish:

        DebugMsg "The program switched to 3GB address space."
        return
endp
end if




; Terminate all process threads
body TerminateAll
begin
        mov     eax, sys_exit_group
        mov     ebx, [.exit_code]
        int     $80
        return
endp




body SetForcedTerminateHandler

  .action lnx_sigaction

begin
        pushad

        mov     eax, [.procHandler]
        mov     [__UserSignalsProc], eax

        mov     [.action.sa_handler], __TerminateSignalsHandler
        mov     [.action.sa_sigaction], 0
        mov     [.action.sa_mask], 0
        mov     [.action.sa_flags], 0
        mov     [.action.sa_restorer], 0

        mov     eax, sys_sigaction
        mov     ebx, SIGTERM
        lea     ecx, [.action]
        xor     edx, edx
        int     $80

        mov     eax, sys_sigaction
        mov     ebx, SIGINT
        lea     ecx, [.action]
        xor     edx, edx
        int     $80

        mov     eax, sys_sigaction
        mov     ebx, SIGQUIT
        lea     ecx, [.action]
        xor     edx, edx
        int     $80

        mov     eax, sys_sigaction
        mov     ebx, SIGHUP
        lea     ecx, [.action]
        xor     edx, edx
        int     $80

        mov     eax, sys_sigaction
        mov     ebx, SIGUSR1
        lea     ecx, [.action]
        xor     edx, edx
        int     $80

        popad
        return
endp


uglobal
  __UserSignalsProc dd ?
  __MemoryFault     dd ?
endg


proc __TerminateSignalsHandler, .signal
begin
        pushad
        DebugMsg "Terminating event received!"

        cmp     [__UserSignalsProc], 0
        je      .finish

        stdcall [__UserSignalsProc], [.signal]

.finish:
        popad
        cret
endp






body GetCmdArguments
begin
        push    ecx edx esi

        mov     esi, [_ptrEnvironment]

        mov     ecx, [esi]
        add     esi, 4

        stdcall CreateArray, 4
        mov     edx, eax
        jecxz   .array_ok

        stdcall AddArrayItems, eax, ecx

.array_ok:
        push    edx
        jecxz   .arg_ok

        lea     edx, [edx+TArray.array]

.loop:
        lodsd
        stdcall StrDupMem, eax
        mov     [edx], eax
        add     edx, 4
        loop    .loop

.arg_ok:
        pop     eax
        pop     esi edx ecx
        return
endp







proc SetSegmentationFaultHandler, .procHandler

  .action lnx_sigaction

begin
        pushad

        mov     eax, [.procHandler]
        mov     [__MemoryFault], eax

        mov     [.action.sa_handler], __ExceptionHandler
        mov     [.action.sa_sigaction], 0
        mov     [.action.sa_mask], 0
        mov     [.action.sa_flags], 0
        mov     [.action.sa_restorer], 0

        mov     eax, sys_sigaction
        mov     ebx, SIGSEGV
        lea     ecx, [.action]
        xor     edx, edx
        int     $80

        popad
        return
endp



if used __ExceptionHandler

__reg_names db  'eax', sig_context.eax
            db  'ebx', sig_context.ebx
            db  'ecx', sig_context.ecx
            db  'edx', sig_context.edx
            db  'esi', sig_context.esi
            db  'edi', sig_context.edi
            db  'ebp', sig_context.ebp
            db  'esp', sig_context.esp
            dd  0

__ExceptionHandler:

        lea     esi, [esp+8]    ; the start of sig_context structure.

        stdcall FileOpenAccess, "fault.txt", faCreateAlways or faWriteOnly
        mov     ebx, eax

; Write the registers to the file:

        mov     edi, __reg_names

.reg_loop:
        cmp     byte [edi], 0
        je      .dump_stack

        stdcall FileWrite, ebx, edi, 3
        stdcall FileWrite, ebx, txt " = ", 3

        movzx   eax, byte [edi+3]
        stdcall NumToStr, [esi + eax], ntsHex or ntsUnsigned or ntsFixedWidth + 8
        stdcall FileWriteString, ebx, eax
        stdcall FileWriteString, ebx, <txt 13, 10>
        add     edi, 4
        jmp     .reg_loop

; Write the stack backtrace to the file:

.dump_stack:
        stdcall FileWriteString, ebx, <txt 13, 10>
        mov     eax, [esi+sig_context.eip]

        stdcall NumToStr, eax, ntsHex or ntsUnsigned or ntsFixedWidth + 8
        stdcall FileWriteString, ebx, eax
        stdcall FileWriteString, ebx, <txt 13, 10, 13, 10>

        lea     ebp, [esi+sig_context.ebp]

.loop:
        mov     ebp, [ebp]
        mov     eax, [ebp+4]    ; return address

        stdcall NumToStr, eax, ntsHex or ntsUnsigned or ntsFixedWidth + 8
        stdcall FileWriteString, ebx, eax
        stdcall FileWriteString, ebx, <txt 13, 10>

        cmp     dword [ebp+4], $b0000000
        jb      .loop

.finish:
        stdcall FileClose, ebx

        cmp     [__MemoryFault], 0
        je      .return

        jmp     [__MemoryFault]

.return:
        retn


end if



body Exec2
.pArgs dd ?
begin
        pushad

        stdcall StrSplitArg, [.hCommand]
        mov     [.pArgs], eax

        mov     eax, sys_fork
        int     $80

        test    eax, eax
        jnz     .parent         ; this is the parent process

; here is the child.

        DebugMsg "Child process here!"

        mov     eax, sys_dup2
        mov     ebx, [.hStdIn]
        xor     ecx, ecx
        int     $80

        mov     eax, sys_dup2
        mov     ebx, [.hStdOut]
        inc     ecx
        int     $80

        mov     eax, sys_dup2
        mov     ebx, [.hStdErr]
        inc     ecx
        int     $80

        mov     ecx, [.pArgs]
        mov     ebx, [ecx]                      ; the first argument is the executable name.

        mov     edx, [_ptrEnvironment]
        mov     eax, [edx]                      ; argument count
        lea     edx, [edx + 4*eax + 8]          ; +4 for argc and +4 for the termination 0; edx points to the env table here.

        mov     eax, sys_execve
        int     $80

        OutputValue "Error Exec2: ", eax, 10, -1

        stc
        mov     [esp+4*regEAX], eax
        popad
        return

.parent:
        stdcall FreeMem, [.pArgs]
        bt      eax, 31
        mov     [esp+4*regEAX], eax
        popad
        return
endp



body WaitProcessExit
begin
        pushad
        mov     eax, sys_waitpid
        mov     ebx, [.ProcessID]
        xor     ecx, ecx
        xor     edx, edx
        int     $80
        clc
        popad
        return
endp



; Splits the string on spaces and forms one-memory-block area that contains, zero terminated
; array of pointers to every part of the string.
; The spaces inside double or single quotes are ignored.
;
; The format of the memory block is the same as used in the sys_execve system function.

proc StrSplitArg, .hCommand
begin
        pushad

if defined StrLib & StrLib
        stdcall StrPtr, [.hCommand]
        mov     esi, eax
        mov     ebx, eax
else
        mov     esi, [.hCommand]
        mov     ebx, esi
end if
        xor     ecx, ecx
        xor     edx, edx

; find the maximal possible count of the arguments (edx)
; and the length of the string.

.loop1:
        lodsb
        test    al, al
        jz      .endstr1
        inc     ecx
        cmp     al, ' '
        jne     .loop1
        inc     edx
        jmp     .loop1

.endstr1:
        mov     esi, ebx

        add     edx, 3
        lea     edx, [4*edx + ecx]
        add     edx, 3
        and     edx, $fffffffc

        stdcall GetMem, edx
        mov     ebx, eax

        sub     edx, ecx
        sub     edx, 4
        and     edx, $fffffffc

        lea     edi, [ebx+edx]
        rep movsb

        mov     edi, ebx
        lea     esi, [ebx+edx]  ; start of the string.

.search_loop:
        mov     [edi], esi
        add     edi, 4

.loop_to_space:
        lodsb
        test    al, al
        jz      .end_of_string

        cmp     al, '"'
        je      .skip_quoted

        cmp     al, "'"
        je      .skip_quoted

        cmp     al, ' '
        jne     .next

        cmp     ah, ' '
        je      .loop_to_space

        mov     byte [esi-1], 0
        jmp     .search_loop

.next:
        mov     ah, al
        jmp     .loop_to_space


.skip_quoted:
        mov     ah, al

.loop_quote:
        lodsb
        test    al, al
        jz      .end_of_string

        cmp     al, ah
        jne     .loop_quote
        jmp     .loop_to_space


.end_of_string:
        mov     [esp+4*regEAX], ebx
        popad
        return
endp






match =LinuxThreads, LinuxThreads {
  define LinuxThreads posix
}

match =native, LinuxThreads {
  include '_process_native.asm'
}

match =posix, LinuxThreads {
  include '_process_posix.asm'
}




