
; text type flags

ttText    = 0      ; usual text
ttStr     = 1      ; string
ttComment = 2      ; comment
ttNum     = 3      ; number
ttChar    = 4      ; symbol character


; syntax color indexes.

clDefaultText   = 0
clSymbol        = 1
clNumber        = 2
clString        = 3
clComment       = 4
clRegister      = 5
clInstruction   = 6
clDirective     = 7
clPreprocessor  = 8

SyntaxColorCount = 8 ; 1..8 above.

struc synList minLen, maxLen, ptrlist {
  .MinLen  dd minLen
  .MaxLen  dd maxLen
  .ptrList dd ptrlist
}

virtual at 0
  synList synList 0, 0, 0
end virtual


uglobal
  pUserSyntax    dd ?        ; pointer to THashTree with the user defined keywords.
endg



proc fasm_syntax, .ptrLine, .ptrColors
.status      dd ?
.wlen        dd ?
.w           rw aeLineCharLen
.lastword    dd ?
.Quo         dw ?       ; the last open quote character
.wcnt        dd ?

; Sometimes, AsmEdit calls this procedure with wrong ptrLine.
; In these cases GPF appear.

begin
        pushad

        mov     [.status], ttText
        xor     ebx, ebx
        mov     [.wlen], ebx
        mov     [.wcnt], ebx
        mov     esi, [.ptrLine]
        mov     edi, [.ptrColors]
        mov     [.lastword], edi
        mov     ecx, aeLineCharLen

        cmp     esi, $ffff         ; sometimes the AsmEdit passes invalid pointer to syntax and it crashes. CHECK THIS!!!
        jbe     .end

.loop:
        lodsw

        cmp     [.status], ttComment
        je      .endif

        cmp     [.status], ttStr
        jne     .ifnotstr

        cmp     ax, [.Quo]
        jne     .endif

        mov     [.status], ttText
        jmp     .endif

.ifnotstr:
        cmp     ax, ';'
        jne     .notcomment

        mov     [.status], ttComment
        mov     bl, clComment

        call    .ProcessWord
        jmp     .endif

.notcomment:
        cmp     ax, '"'
        je      .str
        cmp     ax, "'"
        jne     .notstr

.str:
        mov     [.status], ttStr
        mov     [.Quo], ax
        mov     bl, clString

        call    .ProcessWord
        jmp     .endif

.notstr:
        test    ah, ah
        jnz     .notsym

        call    .CharInSymbols
        jnc     .notsym

        mov     [.status], ttChar
        mov     bl, clSymbol

        call    .ProcessWord
        jmp     .endif

.notsym:
        lea     edx, [esi-aeCharSize]
        cmp     edx, [.ptrLine]
        je      .start_word

        cmp     [.status], ttChar
        jne     .not_word_start

.start_word:
        mov     [.status], ttText

        xor     ebx, ebx
        mov     [.wlen], aeCharSize
        mov     [.w], ax
        mov     [.lastword], edi
        jmp     .endif

.not_word_start:
        cmp     [.status], ttText
        jne     .endif

        lea     ebx, [.w]
        add     ebx, [.wlen]
        mov     [ebx], ax

        xor     ebx, ebx
        add     [.wlen], aeCharSize

.endif:
        mov     [edi], bx
        lea     edi, [edi + aeCharSize]
        dec     ecx
        jnz     .loop

        popad
        return


.ProcessWord:
        pushad

        mov     eax, [.wlen]
        test    eax, eax
        jz      .end

        mov     word [.w + eax], 0

        lea     esi, [.w]

        stdcall WideCharToUtf8, esi
        push    eax
        stdcall SearchHashTree, [pUserSyntax], eax, FALSE
        stdcall StrDel ; from the stack
        jnc     .user_syntax_ok

        add     eax, [pUserSyntax]
        mov     bl, byte [eax+THashTreeLeaf.lparam]
        jmp     .found

.user_syntax_ok:

        push    esi edi

        lea     edi, [.w]

 .new_char:
        lodsw
        cmp     ax, 0
        je      .ok

        test    ah, ah
        jnz     @f

        cmp     ax,"A"
        jl      @F
        cmp     ax,"Z"
        jg      @F
        add     ax, 20h         ; lower case

@@:
        stosw
        jmp     .new_char

.ok:
        inc     [.wcnt]
        pop     edi esi

        mov     bl, clNumber
        call    .IsNumber
        jc      .found

        mov     bl, clRegister

        stdcall FindInstr, sSymbols, esi, [.wlen], 2, 0
        jnc     .maybeinstr

        cmp     dl, 10h
        je      .found
        cmp     dl, 14h
        je      .found

        mov     bl, clDirective
        jmp     .found

.maybeinstr:
        mov     bl, clInstruction

        stdcall FindInstr, sInstructions, esi, [.wlen], 3, 1
        jnc     .maybeoperator

        cmp     [.wcnt], 2
        ja      .directive

        cmp     edx, fasm_instructions_start - instruction_handler
        jb      .directive

        cmp     edx, fasm_instructions_end - instruction_handler
        jbe     .found

.directive:
        mov     bl, clDirective
        jmp     .found

.maybeoperator:
        mov     bl, clDirective
        stdcall SearchSimple, directive_operators, esi, [.wlen], 1
        jc      .found

        stdcall FindInstr, sDataDirectives, esi, [.wlen], 3, -1
        jc      .found

        mov     bl, clPreprocessor
        stdcall SearchSimple, preprocessor_directives, esi, [.wlen], 2
        jc      .found

        stdcall SearchSimple, macro_directives, esi, [.wlen], 2
        jnc     .notfound

.found:
        mov     ecx, [.wlen]
        mov     edi, [.lastword]

.chgcolor:
        mov     [edi], bx
        lea     edi, [edi+aeCharSize]

        dec     ecx
        jnz     .chgcolor

.notfound:
        and     [.wlen], 0

.end:
        popad
        retn


.CharInSymbols:
        push    edi ecx

        mov     edi, symbol_characters+1
        movzx   ecx, byte [edi-1]

        repne scasb
        jne     @f

        stc
        pop     ecx edi
        ret
@@:
        clc
        pop     ecx edi
        retn

; dl = 0 for decimal
; dl = 1 for hex
; dh: number of '.' characters.
; cl: max character
; ch: number of 'e' characters.
; cl: number of hex digits
; bl: the max of all digits.

.IsNumber:
        push    esi ebx

        xor     ebx, ebx
        xor     edx, edx
        xor     ecx, ecx

        lodsw

        cmp     ax, '$'
        je      .hex_num

        cmp     ax, '0'
        je      .maybe_hex
        jb      .not_num
        cmp     ax, '9'
        jbe     .check_max
        jmp     .not_num

.maybe_hex:
        mov     ax, [esi]
        test    ax, ax
        jz      .yes_num

        cmp     ax, 'x'
        je      .hex_num2

        mov     bl, '0'
        jmp     .num_loop

.hex_num2:
        add     esi, aeCharSize

.hex_num:
        mov     dl, 1
        jmp     .num_loop

.check_max:
        cmp     bl, al
        jae     .num_loop
        mov     bl, al

.num_loop:
        lodsw
        test    ax, ax
        jz      .check_final

        cmp     ax, '_'
        je      .num_loop

        cmp     ax, '.'
        je      .point

        cmp     ax, '0'
        jb      .not_num
        cmp     ax, '9'
        jbe     .check_max

        cmp     ax, 'o'
        je      .is_last
        cmp     ax, 'h'
        je      .is_last

        cmp     ax, 'a'
        jb      .not_num
        cmp     ax, 'f'
        je      .is_last2
        ja      .not_num

        cmp     ax, 'b'
        je      .is_last3

.inc_hex:
        inc     cl

        cmp     ax, 'e'
        jne     .check_max

        inc     ch
        jmp     .check_max

.point:
        inc     dh
        jmp     .num_loop

.is_last3:
        cmp     word [esi], 0
        jne     .inc_hex

        test    dl, dl
        jnz     .no_float
        jmp     .only_binary_no_float

.is_last2:
        cmp     word [esi], 0
        jne     .inc_hex

        test    dl, dl
        jnz     .no_float
        jmp     .only_decimal

; only for o and h suffix:
.is_last:
        cmp     word [esi], 0
        jne     .not_num

        cmp     ax, 'o'
        je      .only_octal_no_float

; suffix hex here
        test    dl, dl
        jnz     .not_num
        jmp     .no_float

.check_final:
        test    dl, dl
        jz      .only_decimal

.no_float:  ; hex constant
        cmp     dh, 0           ; '.' characters
        jne     .not_num
        test    bl, bl
        jz      .not_num
        jmp     .yes_num

.only_binary_no_float:
        cmp     dh, 0           ; '.' characters
        jne     .not_num
        cmp     ch, 0           ; 'e' characters
        jne     .not_num
        test    bl, bl
        jz      .not_num
        cmp     bl, '1'
        ja      .not_num
        jmp     .yes_num

.only_octal_no_float:

        cmp     dh, 0           ; '.' characters
        jne     .not_num
        cmp     ch, 0           ; 'e' characters
        jne     .not_num
        cmp     cl, 0           ; hex characters
        ja      .not_num
        test    bl,bl
        jz      .not_num
        cmp     bl, '8'
        jae     .not_num
        jmp     .yes_num

.only_decimal:

        cmp     dh, 1           ; '.'
        ja      .not_num
        cmp     ch, 1           ; 'e'
        ja      .not_num
        cmp     cl, ch          ; the number of hex should be equal to the number of 'e' chars.
        je      .yes_num

.not_num:
        pop     ebx esi
        clc
        retn


.yes_num:
        pop     ebx esi
        stc
        retn

endp




proc FindInstr, .ptrArray, .ptrWord, .len, .addbytes, .return
.count dd ?
begin

        push    eax ebx ecx esi edi

        sar     [.len], 1

        mov     ecx, [.ptrArray]
        mov     edi, [.ptrWord]
        mov     edx, [.len]

        cmp     edx, [ecx+synList.MinLen]
        jb      .notfound

        cmp     edx, [ecx+synList.MaxLen]
        ja      .notfound

        mov     esi, [ecx+synList.ptrList]

        sub     edx, [ecx+synList.MinLen]
        movzx   eax, word [esi+4*edx+2]       ; items count
        mov     [.count], eax
        movzx   ebx, word [esi+4*edx]         ; offset to esi
        add     esi, ebx
        mov     edx, [.len]
        add     edx, [.addbytes]

.loopmain:
        dec     [.count]
        js      .notfound

        xor     ebx, ebx
        xor     eax, eax

.loopword:
        mov     ax, [edi+2*ebx]
        cmp     al, [esi+ebx]
        ja      .next
        jb      .notfound
        inc     ebx
        cmp     ebx, [.len]
        jne     .loopword

        cmp     [.return], 0
        jl      .found

        lea     eax, [esi+ebx]
        add     eax, [.return]

        movzx   edx, word [eax]

.found:
        stc
        jmp     .end

.next:
        add     esi, edx
        jmp     .loopmain

.notfound:
        clc

.end:
        pop     edi esi ecx ebx eax
        return
endp



proc SearchSimple, .ptrList, .ptrWord, .len, .addbytes
.count dd ?
begin
        push    eax ecx edx esi edi

        mov     esi, [.ptrList]

        shr     [.len], 1
        mov     edx, [.len]

.loopmain:
        lodsb
        test    al, al
        jz      .notfound

        cmp     al, dl
        jne     .next

        mov     edi, [.ptrWord]
        movzx   ecx, al

        xor     eax, eax

.cmp:
        lodsb
        cmp     [edi], ax
        lea     edi, [edi+2]
        jne     .nxt

        loop    .cmp
        jmp     .found

.nxt:
        lea     esi, [esi+ecx-1]
        add     esi, [.addbytes]
        jmp     .loopmain

.next:
        movzx   eax, al
        lea     esi, [esi+eax]
        add     esi, [.addbytes]
        jmp     .loopmain

.found:
        stc
        jmp     .end

.notfound:
        clc
.end:
        pop     edi esi edx ecx eax
        return
endp





iglobal
  sInstructions   synList 2, 16, instructions
  sSymbols        synList 1, 11, symbols
  sDataDirectives synList 2, 4, data_directives
endg




proc RebuildUserSyntax
begin
        pushad

        cmp     [pUserSyntax], 0
        je      .create_new

        stdcall FreeHashTree, [pUserSyntax]

.create_new:

        stdcall CreateArray, sizeof.THashTreeNode
        mov     [pUserSyntax], eax

        stdcall CfgGetStr, [hCfgFileName], 'UserSyntax', 'Registers'
        jc      @f

        stdcall AddToUserSyntax, eax, clRegister
        stdcall StrDel, eax

@@:
        stdcall CfgGetStr, [hCfgFileName], 'UserSyntax', 'Instructions'
        jc      @f

        stdcall AddToUserSyntax, eax, clInstruction
        stdcall StrDel, eax

@@:
        stdcall CfgGetStr, [hCfgFileName], 'UserSyntax', 'AsmDirectives'
        jc      @f

        stdcall AddToUserSyntax, eax, clDirective
        stdcall StrDel, eax

@@:
        stdcall CfgGetStr, [hCfgFileName], 'UserSyntax', 'PrepDirectives'
        jc      @f

        stdcall AddToUserSyntax, eax, clPreprocessor
        stdcall StrDel, eax

@@:
        popad
        return
endp



proc AddToUserSyntax, .hCommaSeparated, .color
begin
        pushad

        stdcall StrSplitList, [.hCommaSeparated], ',', FALSE
        mov     edi, eax

        xor     ecx, ecx
        dec     ecx

.loop:
        inc     ecx
        cmp     ecx, [edi+TArray.count]
        jae     .end_loop

        stdcall SearchHashTree, [pUserSyntax], [edi+TArray.array+4*ecx], TRUE
        mov     [pUserSyntax], edx
        jc      .loop

        test    eax, eax
        jz      .end_loop ;out of memory error!

        and     [edi+TArray.array+4*ecx], 0
        mov     edx, [.color]
        mov     [eax+THashTreeLeaf.lparam], edx
        jmp     .loop

.end_loop:

        stdcall ListFree, edi, StrDel

        popad
        return
endp